"""OpenTelemetry integration tests for `llm.Model.stream`."""

from __future__ import annotations

from collections.abc import Generator, Iterator, Mapping
from typing import TypedDict
from unittest.mock import Mock

import httpx
import openai
import pytest
from inline_snapshot import snapshot
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export.in_memory_span_exporter import InMemorySpanExporter
from pydantic import BaseModel, Field

from mirascope import llm, ops
from mirascope.ops._internal.configuration import set_tracer
from tests.ops.utils import span_snapshot


@pytest.fixture(autouse=True, scope="function")
def initialize(tracer_provider: TracerProvider) -> Generator[None, None, None]:
    """Initialize ops configuration and LLM instrumentation for each test."""
    ops.configure(tracer_provider=tracer_provider)
    ops.instrument_llm()
    yield
    ops.uninstrument_llm()


def _math_messages() -> list[llm.Message]:
    """Return a simple math prompt for streaming tests."""
    return [llm.messages.user("What is 4200 + 42?")]


@pytest.mark.vcr()
def test_model_stream_exports_genai_span(
    span_exporter: InMemorySpanExporter,
) -> None:
    """Test that streaming a model call exports the correct OpenTelemetry span."""
    ops.instrument_llm()
    model = llm.Model(model_id="openai/gpt-4o")

    response = model.stream(_math_messages())
    response.finish()

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 1


@pytest.mark.vcr()
def test_model_stream_with_tools(span_exporter: InMemorySpanExporter) -> None:
    """Test that streaming a model call exports the correct OpenTelemetry span."""
    ops.instrument_llm()

    @llm.tool
    def secret_retrieval_tool(password: str) -> str:
        """A tool that requires a password to retrieve a secret."""

        return f"Retrieved secret for {password}"

    model = llm.Model(model_id="openai/gpt-4o")
    messages = [
        llm.messages.system("Use parallel tool calling."),
        llm.messages.user(
            "Please retrieve the secrets associated with each of these passwords: mellon,radiance"
        ),
    ]

    response = model.stream(messages, tools=[secret_retrieval_tool])
    response.finish()

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 1
    span_dict = span_snapshot(spans[0])
    assert span_dict == snapshot(
        {
            "name": "chat openai/gpt-4o",
            "kind": "CLIENT",
            "status": "UNSET",
            "attributes": {
                "gen_ai.operation.name": "chat",
                "gen_ai.provider.name": "openai",
                "gen_ai.request.model": "openai/gpt-4o",
                "gen_ai.output.type": "text",
                "gen_ai.tool.definitions": '[{"name":"secret_retrieval_tool","description":"A tool that requires a password to retrieve a secret.","type":"function","parameters":{"properties":{"password":{"title":"Password","type":"string"}},"required":["password"],"additionalProperties":false,"$defs":null}}]',
                "gen_ai.response.model": "openai/gpt-4o",
                "gen_ai.response.finish_reasons": ["stop"],
                "gen_ai.system_instructions": '[{"type":"text","content":"Use parallel tool calling."}]',
                "gen_ai.input.messages": '[{"role":"user","parts":[{"type":"text","content":"Please retrieve the secrets associated with each of these passwords: mellon,radiance"}]}]',
                "gen_ai.output.messages": '[{"role":"assistant","parts":[{"type":"tool_call","id":"call_nQBHB7aQgltW1QH94YRQYyer","name":"secret_retrieval_tool","arguments":{"password":"mellon"}},{"type":"tool_call","id":"call_rI7qIauOBe4l1rEe4enUzA2a","name":"secret_retrieval_tool","arguments":{"password":"radiance"}}],"finish_reason":"stop"}]',
                "gen_ai.usage.input_tokens": 36,
                "gen_ai.usage.output_tokens": 54,
                "mirascope.response.messages": '[{"role":"system","content":{"type":"text","text":"Use parallel tool calling."}},{"role":"user","content":[{"type":"text","text":"Please retrieve the secrets associated with each of these passwords: mellon,radiance"}],"name":null}]',
                "mirascope.response.content": '[{"type":"tool_call","id":"call_nQBHB7aQgltW1QH94YRQYyer","name":"secret_retrieval_tool","args":"{\\"password\\":\\"mellon\\"}"},{"type":"tool_call","id":"call_rI7qIauOBe4l1rEe4enUzA2a","name":"secret_retrieval_tool","args":"{\\"password\\":\\"radiance\\"}"}]',
                "mirascope.response.usage": '{"input_tokens":36,"output_tokens":54,"cache_read_tokens":0,"cache_write_tokens":0,"reasoning_tokens":0,"total_tokens":90}',
            },
        }
    )


@pytest.mark.vcr()
def test_model_stream_with_json_format(
    span_exporter: InMemorySpanExporter,
) -> None:
    """Test that streaming a model call exports the correct OpenTelemetry span."""

    class Author(BaseModel):
        """The author of a book."""

        first_name: str
        last_name: str

    class Book(BaseModel):
        """A book with a rating. The title should be in all caps!"""

        title: str
        author: Author
        rating: int = Field(description="For testing purposes, the rating should be 7")

    ops.instrument_llm()
    model = llm.Model(model_id="openai/gpt-4o")
    messages = [
        llm.messages.system(
            "Always respond to the user's query using the __mirascope_formatted_output_tool__ tool for structured output."
        ),
        llm.messages.user("Please recommend the most popular book by Patrick Rothfuss"),
    ]

    response_format = llm.format(Book, mode="tool")
    response = model.stream(messages, format=response_format)
    response.finish()

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 1
    span_dict = span_snapshot(spans[0])
    assert span_dict == snapshot(
        {
            "name": "chat openai/gpt-4o",
            "kind": "CLIENT",
            "status": "UNSET",
            "attributes": {
                "gen_ai.operation.name": "chat",
                "gen_ai.provider.name": "openai",
                "gen_ai.request.model": "openai/gpt-4o",
                "gen_ai.output.type": "json",
                "gen_ai.tool.definitions": '[{"name":"__mirascope_formatted_output_tool__","description":"Use this tool to extract data in Book format for a final response.\\nA book with a rating. The title should be in all caps!","type":"function","parameters":{"properties":{"title":{"title":"Title","type":"string"},"author":{"$ref":"#/$defs/Author"},"rating":{"description":"For testing purposes, the rating should be 7","title":"Rating","type":"integer"}},"required":["title","author","rating"],"additionalProperties":false,"$defs":{"Author":{"description":"The author of a book.","properties":{"first_name":{"title":"First Name","type":"string"},"last_name":{"title":"Last Name","type":"string"}},"required":["first_name","last_name"],"title":"Author","type":"object"}}}}]',
                "gen_ai.response.model": "openai/gpt-4o",
                "gen_ai.response.finish_reasons": ["stop"],
                "gen_ai.system_instructions": '[{"type":"text","content":"Always respond to the user\'s query using the __mirascope_formatted_output_tool__ tool for structured output."}]',
                "gen_ai.input.messages": '[{"role":"user","parts":[{"type":"text","content":"Please recommend the most popular book by Patrick Rothfuss"}]}]',
                "gen_ai.output.messages": '[{"role":"assistant","parts":[{"type":"text","content":"{\\"title\\":\\"THE NAME OF THE WIND\\",\\"author\\":{\\"first_name\\":\\"Patrick\\",\\"last_name\\":\\"Rothfuss\\"},\\"rating\\":7}"}],"finish_reason":"stop"}]',
                "gen_ai.usage.input_tokens": 170,
                "gen_ai.usage.output_tokens": 30,
                "mirascope.response.messages": '[{"role":"system","content":{"type":"text","text":"Always respond to the user\'s query using the __mirascope_formatted_output_tool__ tool for structured output."}},{"role":"user","content":[{"type":"text","text":"Please recommend the most popular book by Patrick Rothfuss"}],"name":null}]',
                "mirascope.response.content": '[{"type":"text","text":"{\\"title\\":\\"THE NAME OF THE WIND\\",\\"author\\":{\\"first_name\\":\\"Patrick\\",\\"last_name\\":\\"Rothfuss\\"},\\"rating\\":7}"}]',
                "mirascope.response.usage": '{"input_tokens":170,"output_tokens":30,"cache_read_tokens":0,"cache_write_tokens":0,"reasoning_tokens":0,"total_tokens":200}',
            },
        }
    )


@pytest.mark.vcr()
def test_model_stream_records_untracked_params_event(
    span_exporter: InMemorySpanExporter,
) -> None:
    """Test that streaming a model call exports the correct OpenTelemetry span."""
    ops.instrument_llm()

    class _NonSerializable:
        """Dummy object that stringifies cleanly for untracked params tests."""

        def __str__(self) -> str:
            return "custom-param"

    class _BadStr:
        """Dummy object whose __str__ intentionally raises."""

        def __str__(self) -> str:  # pragma: no cover - forced failure path
            raise ValueError("cannot stringify")

    class _TraceMetadata(TypedDict):
        """Trace metadata payload."""

        id: str
        tags: list[str]

    class _Metadata(TypedDict):
        """Metadata container for trace info."""

        trace: _TraceMetadata

    class _ExtraParams(llm.Params, total=False):
        """Params with unsupported and serializable extras for testing."""

        metadata: _Metadata
        unsupported_list: list[int]
        non_serializable: _NonSerializable
        bad_str: _BadStr

    extra_params: _ExtraParams = {
        "metadata": {"trace": {"id": "abc123", "tags": ["otel"]}},
        "unsupported_list": [1, 2, 3],
        "non_serializable": _NonSerializable(),
        "bad_str": _BadStr(),
    }
    model = llm.Model(
        model_id="openai/gpt-4o",
        **extra_params,
    )

    response = model.stream(_math_messages())
    response.finish()

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 1
    events = spans[0].events
    event = events[-1]
    event_attributes: Mapping[str, object] = event.attributes or {}
    event_snapshot = {
        "name": event.name,
        "attributes": {
            key: list(value) if isinstance(value, tuple) else value
            for key, value in event_attributes.items()
        },
    }
    assert event_snapshot == snapshot(
        {
            "name": "gen_ai.request.params.untracked",
            "attributes": {
                "gen_ai.untracked_params.count": 4,
                "gen_ai.untracked_params.keys": [
                    "metadata",
                    "unsupported_list",
                    "non_serializable",
                    "bad_str",
                ],
                "gen_ai.untracked_params.json": '{"metadata":{"trace":{"id":"abc123","tags":["otel"]}},"unsupported_list":[1,2,3],"non_serializable":"custom-param","bad_str":"<_BadStr>"}',
            },
        }
    )


@pytest.mark.vcr()
def test_model_stream_with_none_parameters(
    span_exporter: InMemorySpanExporter,
) -> None:
    """Test that streaming a model call exports the correct OpenTelemetry span."""
    ops.instrument_llm()
    model = llm.Model(
        model_id="openai/gpt-4o",
        temperature=None,  # type: ignore[arg-type]
        max_tokens=64,
        top_p=None,  # type: ignore[arg-type]
    )

    response = model.stream(_math_messages())
    response.finish()

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 1
    span_dict = span_snapshot(spans[0])
    assert span_dict == snapshot(
        {
            "name": "chat openai/gpt-4o",
            "kind": "CLIENT",
            "status": "UNSET",
            "attributes": {
                "gen_ai.operation.name": "chat",
                "gen_ai.provider.name": "openai",
                "gen_ai.request.model": "openai/gpt-4o",
                "gen_ai.output.type": "text",
                "gen_ai.request.max_tokens": 64,
                "gen_ai.response.model": "openai/gpt-4o",
                "gen_ai.response.finish_reasons": ["stop"],
                "gen_ai.input.messages": '[{"role":"user","parts":[{"type":"text","content":"What is 4200 + 42?"}]}]',
                "gen_ai.output.messages": '[{"role":"assistant","parts":[{"type":"text","content":"4200 + 42 = 4242"}],"finish_reason":"stop"}]',
                "gen_ai.usage.input_tokens": 16,
                "gen_ai.usage.output_tokens": 10,
                "mirascope.response.messages": '[{"role":"user","content":[{"type":"text","text":"What is 4200 + 42?"}],"name":null}]',
                "mirascope.response.content": '[{"type":"text","text":"4200 + 42 = 4242"}]',
                "mirascope.response.usage": '{"input_tokens":16,"output_tokens":10,"cache_read_tokens":0,"cache_write_tokens":0,"reasoning_tokens":0,"total_tokens":26}',
            },
        }
    )


@pytest.mark.vcr()
def test_model_stream_records_response_id(
    span_exporter: InMemorySpanExporter,
) -> None:
    """Test that streaming a model call exports the correct OpenTelemetry span."""
    ops.instrument_llm()
    model = llm.Model(model_id="openai/gpt-4o")

    response = model.stream(_math_messages())
    response.finish()

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 1
    attrs = spans[0].attributes or {}
    assert "gen_ai.response.id" not in attrs
    first_span_dict = span_snapshot(spans[0])
    assert first_span_dict == snapshot(
        {
            "name": "chat openai/gpt-4o",
            "kind": "CLIENT",
            "status": "UNSET",
            "attributes": {
                "gen_ai.operation.name": "chat",
                "gen_ai.provider.name": "openai",
                "gen_ai.request.model": "openai/gpt-4o",
                "gen_ai.output.type": "text",
                "gen_ai.response.model": "openai/gpt-4o",
                "gen_ai.response.finish_reasons": ["stop"],
                "gen_ai.input.messages": '[{"role":"user","parts":[{"type":"text","content":"What is 4200 + 42?"}]}]',
                "gen_ai.output.messages": '[{"role":"assistant","parts":[{"type":"text","content":"4200 + 42 equals 4242."}],"finish_reason":"stop"}]',
                "gen_ai.usage.input_tokens": 16,
                "gen_ai.usage.output_tokens": 11,
                "mirascope.response.messages": '[{"role":"user","content":[{"type":"text","text":"What is 4200 + 42?"}],"name":null}]',
                "mirascope.response.content": '[{"type":"text","text":"4200 + 42 equals 4242."}]',
                "mirascope.response.usage": '{"input_tokens":16,"output_tokens":11,"cache_read_tokens":0,"cache_write_tokens":0,"reasoning_tokens":0,"total_tokens":27}',
            },
        }
    )


@pytest.mark.vcr()
def test_model_stream_without_instrumentation(
    span_exporter: InMemorySpanExporter,
) -> None:
    """Test that streaming a model call exports the correct OpenTelemetry span."""
    ops.uninstrument_llm()

    model = llm.Model(model_id="openai/gpt-4o")
    response = model.stream(_math_messages())
    response.finish()

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 0


@pytest.mark.vcr()
def test_model_stream_with_tracer_set_to_none(
    span_exporter: InMemorySpanExporter,
) -> None:
    ops.instrument_llm()
    set_tracer(None)

    model = llm.Model(model_id="openai/gpt-4o")
    response = model.stream(_math_messages())
    response.finish()

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 0


def test_model_stream_with_error(span_exporter: InMemorySpanExporter) -> None:
    """Test that streaming a model call exports the correct OpenTelemetry span."""
    ops.instrument_llm()

    # Create a mock provider that raises an error
    mock_provider = Mock()
    mock_provider.id = "openai"
    mock_provider.error_map = {}  # Empty error map - errors pass through
    mock_provider.stream.side_effect = openai.APIError(
        "Server error occurred",
        request=httpx.Request("POST", "https://example.com"),
        body=None,
    )

    # Register the broken provider
    llm.register_provider(mock_provider, scope="openai/")

    model = llm.Model(model_id="openai/gpt-4o")
    with pytest.raises(openai.APIError):
        model.stream(_math_messages())

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 1
    span_dict = span_snapshot(spans[0])
    assert span_dict == snapshot(
        {
            "name": "chat openai/gpt-4o",
            "kind": "CLIENT",
            "status": "ERROR",
            "attributes": {
                "gen_ai.operation.name": "chat",
                "gen_ai.provider.name": "openai",
                "gen_ai.request.model": "openai/gpt-4o",
                "gen_ai.output.type": "text",
                "gen_ai.input.messages": '[{"role":"user","parts":[{"type":"text","content":"What is 4200 + 42?"}]}]',
                "error.type": "APIError",
                "error.message": "Server error occurred",
            },
        }
    )


def test_model_stream_iterator_error_records_exception(
    span_exporter: InMemorySpanExporter,
) -> None:
    """Test that streaming a model call exports the correct OpenTelemetry span."""
    ops.instrument_llm()

    def _chunk_iterator() -> Iterator[llm.StreamResponseChunk]:
        """Return a chunk iterator that raises to simulate stream failure."""
        yield llm.TextStartChunk()
        yield llm.TextChunk(delta="partial")
        raise RuntimeError("chunk boom")

    # Create a mock provider that returns a failing stream
    mock_provider = Mock()
    mock_provider.id = "openai"
    mock_provider.error_map = {}  # Empty error map - errors pass through
    mock_provider.stream.return_value = llm.StreamResponse(
        provider_id="openai",
        model_id="openai/gpt-4o",
        provider_model_name="gpt-4o:responses",
        params={},
        tools=None,
        format=None,
        input_messages=_math_messages(),
        chunk_iterator=_chunk_iterator(),
    )

    # Register the mock provider
    llm.register_provider(mock_provider, scope="openai/")

    model = llm.Model(model_id="openai/gpt-4o")
    response = model.stream(_math_messages())
    with pytest.raises(RuntimeError):
        response.finish()

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 1
    span = spans[0]
    assert span.status.status_code.name == "ERROR"
    attrs = span.attributes or {}
    assert attrs["error.type"] == "RuntimeError"
    assert attrs["error.message"] == "chunk boom"



"""End-to-end tests for prompt caching."""

import pytest

from mirascope import llm
from tests.e2e.conftest import E2E_MODEL_IDS
from tests.utils import Snapshot, snapshot_test


@pytest.mark.parametrize("model_id", E2E_MODEL_IDS)
@pytest.mark.vcr
def test_web_search_tool(
    model_id: llm.ModelId,
    snapshot: Snapshot,
) -> None:
    """Test handling of the provider's web search tool."""

    @llm.call(model_id, tools=[llm.WebSearchTool()])
    def crypto_price_lookup() -> str:
        # Assert the date since the models will otherwise complain that they can't predict the future
        return "Jan 1, 2026 is a date in the past. Use the web search tool to lookup the price of bitcoin on that date."

    with snapshot_test(snapshot) as snap:
        response = crypto_price_lookup()
        # Resume just to make sure we handle re-encoding correctly, etc.
        response = response.resume(
            "Please also look up the price of Ethereum on that date"
        )
        snap.set_response(response)


@pytest.mark.parametrize("model_id", E2E_MODEL_IDS)
@pytest.mark.vcr
def test_web_search_tool_stream(
    model_id: llm.ModelId,
    snapshot: Snapshot,
) -> None:
    """Test streaming handling of the provider's web search tool."""

    @llm.call(model_id, tools=[llm.WebSearchTool()])
    def crypto_price_lookup() -> str:
        # Assert the date since the models will otherwise complain that they can't predict the future
        return "Jan 1, 2026 is a date in the past. Use the web search tool to lookup the price of bitcoin on that date."

    with snapshot_test(snapshot) as snap:
        response = crypto_price_lookup.stream()
        response.finish()
        # Resume just to make sure we handle re-encoding correctly, etc.
        response = response.resume(
            "Please also look up the price of Ethereum on that date"
        )
        response.finish()
        snap.set_response(response)



from inline_snapshot import snapshot

from mirascope.llm import (
    AssistantMessage,
    SystemMessage,
    Text,
    ToolCall,
    ToolOutput,
    UserMessage,
)

sync_snapshot = snapshot(
    {
        "response": {
            "provider_id": "openai",
            "model_id": "openai/gpt-4o:completions",
            "provider_model_name": "gpt-4o:completions",
            "params": {},
            "finish_reason": None,
            "usage": {
                "input_tokens": 149,
                "output_tokens": 36,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 0,
                "provider_tool_usage": None,
                "raw": "CompletionUsage(completion_tokens=36, prompt_tokens=149, total_tokens=185, completion_tokens_details=CompletionTokensDetails(accepted_prediction_tokens=0, audio_tokens=0, reasoning_tokens=0, rejected_prediction_tokens=0), prompt_tokens_details=PromptTokensDetails(audio_tokens=0, cached_tokens=0))",
                "total_tokens": 185,
            },
            "messages": [
                SystemMessage(content=Text(text="Use parallel tool calling.")),
                UserMessage(
                    content=[
                        Text(
                            text="Please retrieve the secrets associated with each of these passwords: mellon,radiance"
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        ToolCall(
                            id="call_kORjJVpmh91BGUtfuEyWTNHq",
                            name="secret_retrieval_tool",
                            args='{"password": "mellon"}',
                        ),
                        ToolCall(
                            id="call_rwWftDkV0SjAa03c30QMCbHv",
                            name="secret_retrieval_tool",
                            args='{"password": "radiance"}',
                        ),
                    ],
                    provider_id="openai",
                    model_id="openai/gpt-4o:completions",
                    provider_model_name="gpt-4o:completions",
                    raw_message={
                        "role": "assistant",
                        "annotations": [],
                        "tool_calls": [
                            {
                                "id": "call_kORjJVpmh91BGUtfuEyWTNHq",
                                "function": {
                                    "arguments": '{"password": "mellon"}',
                                    "name": "secret_retrieval_tool",
                                },
                                "type": "function",
                            },
                            {
                                "id": "call_rwWftDkV0SjAa03c30QMCbHv",
                                "function": {
                                    "arguments": '{"password": "radiance"}',
                                    "name": "secret_retrieval_tool",
                                },
                                "type": "function",
                            },
                        ],
                    },
                ),
                UserMessage(
                    content=[
                        ToolOutput(
                            id="call_kORjJVpmh91BGUtfuEyWTNHq",
                            name="secret_retrieval_tool",
                            result="Welcome to Moria!",
                        ),
                        ToolOutput(
                            id="call_rwWftDkV0SjAa03c30QMCbHv",
                            name="secret_retrieval_tool",
                            result="Life before Death",
                        ),
                    ]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text="""\
Here are the secrets associated with the passwords you provided:

- For "mellon": "Welcome to Moria!"
- For "radiance": "Life before Death"\
"""
                        )
                    ],
                    provider_id="openai",
                    model_id="openai/gpt-4o:completions",
                    provider_model_name="gpt-4o:completions",
                    raw_message={
                        "content": """\
Here are the secrets associated with the passwords you provided:

- For "mellon": "Welcome to Moria!"
- For "radiance": "Life before Death"\
""",
                        "role": "assistant",
                        "annotations": [],
                    },
                ),
            ],
            "format": None,
            "tools": [
                {
                    "name": "secret_retrieval_tool",
                    "description": "A tool that requires a password to retrieve a secret.",
                    "parameters": """\
{
  "properties": {
    "password": {
      "title": "Password",
      "type": "string"
    }
  },
  "required": [
    "password"
  ],
  "additionalProperties": false,
  "defs": null
}\
""",
                    "strict": None,
                }
            ],
        }
    }
)
async_snapshot = snapshot(
    {
        "response": {
            "provider_id": "openai",
            "model_id": "openai/gpt-4o:completions",
            "provider_model_name": "gpt-4o:completions",
            "params": {},
            "finish_reason": None,
            "usage": {
                "input_tokens": 149,
                "output_tokens": 32,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 0,
                "provider_tool_usage": None,
                "raw": "CompletionUsage(completion_tokens=32, prompt_tokens=149, total_tokens=181, completion_tokens_details=CompletionTokensDetails(accepted_prediction_tokens=0, audio_tokens=0, reasoning_tokens=0, rejected_prediction_tokens=0), prompt_tokens_details=PromptTokensDetails(audio_tokens=0, cached_tokens=0))",
                "total_tokens": 181,
            },
            "messages": [
                SystemMessage(content=Text(text="Use parallel tool calling.")),
                UserMessage(
                    content=[
                        Text(
                            text="Please retrieve the secrets associated with each of these passwords: mellon,radiance"
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        ToolCall(
                            id="call_eTrqJEDnTlTGliM458pkKaaQ",
                            name="secret_retrieval_tool",
                            args='{"password": "mellon"}',
                        ),
                        ToolCall(
                            id="call_vgAUPfwzgxIkCV15CYFmASwn",
                            name="secret_retrieval_tool",
                            args='{"password": "radiance"}',
                        ),
                    ],
                    provider_id="openai",
                    model_id="openai/gpt-4o:completions",
                    provider_model_name="gpt-4o:completions",
                    raw_message={
                        "role": "assistant",
                        "annotations": [],
                        "tool_calls": [
                            {
                                "id": "call_eTrqJEDnTlTGliM458pkKaaQ",
                                "function": {
                                    "arguments": '{"password": "mellon"}',
                                    "name": "secret_retrieval_tool",
                                },
                                "type": "function",
                            },
                            {
                                "id": "call_vgAUPfwzgxIkCV15CYFmASwn",
                                "function": {
                                    "arguments": '{"password": "radiance"}',
                                    "name": "secret_retrieval_tool",
                                },
                                "type": "function",
                            },
                        ],
                    },
                ),
                UserMessage(
                    content=[
                        ToolOutput(
                            id="call_eTrqJEDnTlTGliM458pkKaaQ",
                            name="secret_retrieval_tool",
                            result="Welcome to Moria!",
                        ),
                        ToolOutput(
                            id="call_vgAUPfwzgxIkCV15CYFmASwn",
                            name="secret_retrieval_tool",
                            result="Life before Death",
                        ),
                    ]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text="""\
The secrets associated with the passwords are as follows:
- For "mellon": Welcome to Moria!
- For "radiance": Life before Death\
"""
                        )
                    ],
                    provider_id="openai",
                    model_id="openai/gpt-4o:completions",
                    provider_model_name="gpt-4o:completions",
                    raw_message={
                        "content": """\
The secrets associated with the passwords are as follows:
- For "mellon": Welcome to Moria!
- For "radiance": Life before Death\
""",
                        "role": "assistant",
                        "annotations": [],
                    },
                ),
            ],
            "format": None,
            "tools": [
                {
                    "name": "secret_retrieval_tool",
                    "description": "A tool that requires a password to retrieve a secret.",
                    "parameters": """\
{
  "properties": {
    "password": {
      "title": "Password",
      "type": "string"
    }
  },
  "required": [
    "password"
  ],
  "additionalProperties": false,
  "defs": null
}\
""",
                    "strict": None,
                }
            ],
        }
    }
)
stream_snapshot = snapshot(
    {
        "response": {
            "provider_id": "openai",
            "model_id": "openai/gpt-4o:completions",
            "provider_model_name": "gpt-4o:completions",
            "finish_reason": None,
            "messages": [
                SystemMessage(content=Text(text="Use parallel tool calling.")),
                UserMessage(
                    content=[
                        Text(
                            text="Please retrieve the secrets associated with each of these passwords: mellon,radiance"
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        ToolCall(
                            id="call_NnkmjbtYL9gdZnGOMxYEl0A3",
                            name="secret_retrieval_tool",
                            args='{"password": "mellon"}',
                        ),
                        ToolCall(
                            id="call_uvC42eCVUOg2Y35G9FQKjAOB",
                            name="secret_retrieval_tool",
                            args='{"password": "radiance"}',
                        ),
                    ],
                    provider_id="openai",
                    model_id="openai/gpt-4o:completions",
                    provider_model_name="gpt-4o:completions",
                    raw_message=None,
                ),
                UserMessage(
                    content=[
                        ToolOutput(
                            id="call_NnkmjbtYL9gdZnGOMxYEl0A3",
                            name="secret_retrieval_tool",
                            result="Welcome to Moria!",
                        ),
                        ToolOutput(
                            id="call_uvC42eCVUOg2Y35G9FQKjAOB",
                            name="secret_retrieval_tool",
                            result="Life before Death",
                        ),
                    ]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text="""\
The secrets associated with the passwords are as follows:

- "mellon": Welcome to Moria!
- "radiance": Life before Death\
"""
                        )
                    ],
                    provider_id="openai",
                    model_id="openai/gpt-4o:completions",
                    provider_model_name="gpt-4o:completions",
                    raw_message=None,
                ),
            ],
            "format": None,
            "tools": [
                {
                    "name": "secret_retrieval_tool",
                    "description": "A tool that requires a password to retrieve a secret.",
                    "parameters": """\
{
  "properties": {
    "password": {
      "title": "Password",
      "type": "string"
    }
  },
  "required": [
    "password"
  ],
  "additionalProperties": false,
  "defs": null
}\
""",
                    "strict": None,
                }
            ],
            "usage": {
                "input_tokens": 149,
                "output_tokens": 30,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 0,
                "provider_tool_usage": None,
                "raw": "None",
                "total_tokens": 179,
            },
            "n_chunks": 31,
        }
    }
)
async_stream_snapshot = snapshot(
    {
        "response": {
            "provider_id": "openai",
            "model_id": "openai/gpt-4o:completions",
            "provider_model_name": "gpt-4o:completions",
            "finish_reason": None,
            "messages": [
                SystemMessage(content=Text(text="Use parallel tool calling.")),
                UserMessage(
                    content=[
                        Text(
                            text="Please retrieve the secrets associated with each of these passwords: mellon,radiance"
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        ToolCall(
                            id="call_FrKwqXDWJFYAzpMNuKWtQ9PO",
                            name="secret_retrieval_tool",
                            args='{"password": "mellon"}',
                        ),
                        ToolCall(
                            id="call_Z9OUz90nWgDjqOmcMOQK5KbJ",
                            name="secret_retrieval_tool",
                            args='{"password": "radiance"}',
                        ),
                    ],
                    provider_id="openai",
                    model_id="openai/gpt-4o:completions",
                    provider_model_name="gpt-4o:completions",
                    raw_message=None,
                ),
                UserMessage(
                    content=[
                        ToolOutput(
                            id="call_FrKwqXDWJFYAzpMNuKWtQ9PO",
                            name="secret_retrieval_tool",
                            result="Welcome to Moria!",
                        ),
                        ToolOutput(
                            id="call_Z9OUz90nWgDjqOmcMOQK5KbJ",
                            name="secret_retrieval_tool",
                            result="Life before Death",
                        ),
                    ]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text="""\
Here are the retrieved secrets for the passwords:

- For "mellon": Welcome to Moria!
- For "radiance": Life before Death\
"""
                        )
                    ],
                    provider_id="openai",
                    model_id="openai/gpt-4o:completions",
                    provider_model_name="gpt-4o:completions",
                    raw_message=None,
                ),
            ],
            "format": None,
            "tools": [
                {
                    "name": "secret_retrieval_tool",
                    "description": "A tool that requires a password to retrieve a secret.",
                    "parameters": """\
{
  "properties": {
    "password": {
      "title": "Password",
      "type": "string"
    }
  },
  "required": [
    "password"
  ],
  "additionalProperties": false,
  "defs": null
}\
""",
                    "strict": None,
                }
            ],
            "usage": {
                "input_tokens": 149,
                "output_tokens": 31,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 0,
                "provider_tool_usage": None,
                "raw": "None",
                "total_tokens": 180,
            },
            "n_chunks": 32,
        }
    }
)



"""OpenRouter provider implementation."""

from __future__ import annotations

from typing import ClassVar

from ..openai.completions._utils import (
    CompletionsModelFeatureInfo,
    feature_info_for_openai_model,
)
from ..openai.completions.base_provider import BaseOpenAICompletionsProvider
from ..openai.model_id import model_name as openai_model_name


class OpenRouterProvider(BaseOpenAICompletionsProvider):
    """Provider for OpenRouter's OpenAI-compatible API.

    Inherits from BaseOpenAICompletionsProvider with OpenRouter-specific configuration:
    - Uses OpenRouter's API endpoint
    - Requires OPENROUTER_API_KEY

    Usage:
        Option 1: Use "openrouter/" prefix for explicit OpenRouter models:

        ```python
        from mirascope import llm

        llm.register_provider("openrouter", scope="openrouter/")

        @llm.call("openrouter/openai/gpt-4o")
        def my_prompt():
            return [llm.messages.user("Hello!")]
        ```

        Option 2: Route existing model IDs through OpenRouter:

        ```python
        from mirascope import llm

        # Register for openai and anthropic models via OpenRouter
        llm.register_provider("openrouter", scope=["openai/", "anthropic/"])

        # Now openai/ models go through OpenRouter
        @llm.call("openai/gpt-4")
        def my_prompt():
            return [llm.messages.user("Hello!")]
        ```
    """

    id: ClassVar[str] = "openrouter"
    default_scope: ClassVar[str | list[str]] = []
    default_base_url: ClassVar[str | None] = "https://openrouter.ai/api/v1"
    api_key_env_var: ClassVar[str] = "OPENROUTER_API_KEY"
    api_key_required: ClassVar[bool] = True
    provider_name: ClassVar[str | None] = "OpenRouter"

    def _model_name(self, model_id: str) -> str:
        """Strip 'openrouter/' prefix from model ID for OpenRouter API."""
        return model_id.removeprefix("openrouter/")

    def _model_feature_info(self, model_id: str) -> CompletionsModelFeatureInfo:
        """Return OpenAI feature info for openai/* models, empty info otherwise."""
        base_model_id = model_id.removeprefix("openrouter/")
        if base_model_id.startswith("openai/"):
            openai_name = openai_model_name(base_model_id, None)
            return feature_info_for_openai_model(openai_name)
        return CompletionsModelFeatureInfo()



"""RetryPrompt extends Prompt to return RetryResponse instead of Response."""

from collections.abc import Callable
from typing import Any, Generic, overload

from ..formatting import FormatSpec, FormattableT
from ..models import Model
from ..prompts import (
    AsyncPrompt,
    Prompt,
)
from ..prompts.prompts import BasePrompt
from ..prompts.protocols import AsyncMessageTemplate, MessageTemplate
from ..providers import ModelId
from ..tools import AsyncToolkit, Toolkit
from ..types import P
from .retry_config import RetryConfig
from .retry_models import RetryModel
from .retry_responses import (
    AsyncRetryResponse,
    RetryResponse,
)
from .retry_stream_responses import (
    AsyncRetryStreamResponse,
    RetryStreamResponse,
)


def _ensure_retry_model(model: Model | ModelId, config: RetryConfig) -> RetryModel:
    """Ensure a model has retry capabilities, adding them if necessary."""
    if isinstance(model, RetryModel):
        return model
    if isinstance(model, str):
        model = Model(model)
    return RetryModel(model=model, retry_config=config)


class BaseRetryPrompt(BasePrompt[Callable[..., Any]]):
    """Base class for retry-enabled prompts that adds retry_config."""

    retry_config: RetryConfig
    """Configuration for retry behavior."""


class RetryPrompt(BaseRetryPrompt, Prompt[P, FormattableT], Generic[P, FormattableT]):
    """A retry-enabled prompt that extends Prompt and returns RetryResponse.

    This extends Prompt and overrides call methods to return RetryResponse
    instead of Response. It handles wrapping the provided Model in a RetryModel
    if necessary.
    """

    def __init__(
        self,
        *,
        fn: MessageTemplate[P],
        toolkit: Toolkit,
        format: FormatSpec[FormattableT] | None,
        retry_config: RetryConfig,
    ) -> None:
        super().__init__(fn=fn, toolkit=toolkit, format=format)
        self.retry_config = retry_config

    @overload
    def __call__(
        self: "RetryPrompt[P, None]",
        model: Model | ModelId,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> RetryResponse[None]: ...

    @overload
    def __call__(
        self: "RetryPrompt[P, FormattableT]",
        model: Model | ModelId,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> RetryResponse[FormattableT]: ...

    def __call__(
        self, model: Model | ModelId, *args: P.args, **kwargs: P.kwargs
    ) -> RetryResponse[None] | RetryResponse[FormattableT]:
        """Generates a retry response using the provided model."""
        return self.call(model, *args, **kwargs)

    @overload
    def call(
        self: "RetryPrompt[P, None]",
        model: Model | ModelId,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> RetryResponse[None]: ...

    @overload
    def call(
        self: "RetryPrompt[P, FormattableT]",
        model: Model | ModelId,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> RetryResponse[FormattableT]: ...

    def call(
        self, model: Model | ModelId, *args: P.args, **kwargs: P.kwargs
    ) -> RetryResponse[None] | RetryResponse[FormattableT]:
        """Generates a retry response using the provided model."""
        retry_model = _ensure_retry_model(model, self.retry_config)
        messages = self.messages(*args, **kwargs)
        return retry_model.call(
            content=messages, tools=self.toolkit, format=self.format
        )

    @overload
    def stream(
        self: "RetryPrompt[P, None]",
        model: Model | ModelId,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> RetryStreamResponse[None]: ...

    @overload
    def stream(
        self: "RetryPrompt[P, FormattableT]",
        model: Model | ModelId,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> RetryStreamResponse[FormattableT]: ...

    def stream(
        self, model: Model | ModelId, *args: P.args, **kwargs: P.kwargs
    ) -> RetryStreamResponse[None] | RetryStreamResponse[FormattableT]:
        """Generates a retry stream response using the provided model."""
        retry_model = _ensure_retry_model(model, self.retry_config)
        messages = self.messages(*args, **kwargs)
        return retry_model.stream(
            content=messages, tools=self.toolkit, format=self.format
        )


class AsyncRetryPrompt(
    BaseRetryPrompt, AsyncPrompt[P, FormattableT], Generic[P, FormattableT]
):
    """An async retry-enabled prompt that extends AsyncPrompt and returns AsyncRetryResponse."""

    def __init__(
        self,
        *,
        fn: AsyncMessageTemplate[P],
        toolkit: AsyncToolkit,
        format: FormatSpec[FormattableT] | None,
        retry_config: RetryConfig,
    ) -> None:
        super().__init__(fn=fn, toolkit=toolkit, format=format)
        self.retry_config = retry_config

    @overload
    async def __call__(
        self: "AsyncRetryPrompt[P, None]",
        model: Model | ModelId,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> AsyncRetryResponse[None]: ...

    @overload
    async def __call__(
        self: "AsyncRetryPrompt[P, FormattableT]",
        model: Model | ModelId,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> AsyncRetryResponse[FormattableT]: ...

    async def __call__(
        self, model: Model | ModelId, *args: P.args, **kwargs: P.kwargs
    ) -> AsyncRetryResponse[None] | AsyncRetryResponse[FormattableT]:
        """Generates a retry response using the provided model asynchronously."""
        return await self.call(model, *args, **kwargs)

    @overload
    async def call(
        self: "AsyncRetryPrompt[P, None]",
        model: Model | ModelId,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> AsyncRetryResponse[None]: ...

    @overload
    async def call(
        self: "AsyncRetryPrompt[P, FormattableT]",
        model: Model | ModelId,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> AsyncRetryResponse[FormattableT]: ...

    async def call(
        self, model: Model | ModelId, *args: P.args, **kwargs: P.kwargs
    ) -> AsyncRetryResponse[None] | AsyncRetryResponse[FormattableT]:
        """Generates a retry response using the provided model asynchronously."""
        retry_model = _ensure_retry_model(model, self.retry_config)
        messages = await self.messages(*args, **kwargs)
        return await retry_model.call_async(
            content=messages, tools=self.toolkit, format=self.format
        )

    @overload
    async def stream(
        self: "AsyncRetryPrompt[P, None]",
        model: Model | ModelId,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> AsyncRetryStreamResponse[None]: ...

    @overload
    async def stream(
        self: "AsyncRetryPrompt[P, FormattableT]",
        model: Model | ModelId,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> AsyncRetryStreamResponse[FormattableT]: ...

    async def stream(
        self, model: Model | ModelId, *args: P.args, **kwargs: P.kwargs
    ) -> AsyncRetryStreamResponse[None] | AsyncRetryStreamResponse[FormattableT]:
        """Generates a retry stream response using the provided model asynchronously."""
        retry_model = _ensure_retry_model(model, self.retry_config)
        messages = await self.messages(*args, **kwargs)
        return await retry_model.stream_async(
            content=messages, tools=self.toolkit, format=self.format
        )



import asyncio

from mirascope import llm


@llm.call("openai/gpt-5-mini")
async def recommend_book(genre: str):
    return f"Recommend a {genre} book."


async def main():
    response = await recommend_book("fantasy")
    print(response.text())


asyncio.run(main())



from inline_snapshot import snapshot

from mirascope.llm import (
    AssistantMessage,
    SystemMessage,
    Text,
    UserMessage,
)

test_snapshot = snapshot(
    {
        "response": {
            "provider_id": "anthropic",
            "model_id": "anthropic/claude-haiku-4-5",
            "provider_model_name": "claude-haiku-4-5",
            "finish_reason": None,
            "messages": [
                SystemMessage(
                    content=Text(
                        text="Always respond to the user's query using the __mirascope_formatted_output_tool__ tool for structured output."
                    )
                ),
                UserMessage(content=[Text(text="Please recommend a book to me!")]),
                AssistantMessage(
                    content=[
                        Text(
                            text='{"title": "The Midnight Library", "author": "Matt Haig", "themes": ["Life choices","Second chances","Self-discovery","Regret and redemption"], "rating": 5}'
                        )
                    ],
                    provider_id="anthropic",
                    model_id="anthropic/claude-haiku-4-5",
                    provider_model_name="claude-haiku-4-5",
                    raw_message={
                        "role": "assistant",
                        "content": [
                            {
                                "type": "tool_use",
                                "id": "toolu_01VSp9uXC6E9zQBtfJqNhT4B",
                                "name": "__mirascope_formatted_output_tool__",
                                "input": {
                                    "title": "The Midnight Library",
                                    "author": "Matt Haig",
                                    "themes": [
                                        "Life choices",
                                        "Second chances",
                                        "Self-discovery",
                                        "Regret and redemption",
                                    ],
                                    "rating": 5,
                                },
                            }
                        ],
                    },
                ),
            ],
            "format": {
                "name": "BookReview",
                "description": None,
                "schema": {
                    "properties": {
                        "title": {"title": "Title", "type": "string"},
                        "author": {"title": "Author", "type": "string"},
                        "themes": {
                            "items": {"type": "string"},
                            "title": "Themes",
                            "type": "array",
                        },
                        "rating": {"title": "Rating", "type": "integer"},
                    },
                    "required": ["title", "author", "themes", "rating"],
                    "title": "BookReview",
                    "type": "object",
                },
                "mode": "tool",
                "formatting_instructions": "Always respond to the user's query using the __mirascope_formatted_output_tool__ tool for structured output.",
            },
            "tools": [],
            "usage": {
                "input_tokens": 793,
                "output_tokens": 101,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 0,
                "provider_tool_usage": None,
                "raw": "None",
                "total_tokens": 894,
            },
            "n_chunks": 25,
        },
        "structured_stream": [
            {"title": None, "author": None, "themes": None, "rating": None},
            {"title": "T", "author": None, "themes": None, "rating": None},
            {"title": "The Midnig", "author": None, "themes": None, "rating": None},
            {
                "title": "The Midnight Libr",
                "author": None,
                "themes": None,
                "rating": None,
            },
            {
                "title": "The Midnight Library",
                "author": None,
                "themes": None,
                "rating": None,
            },
            {
                "title": "The Midnight Library",
                "author": None,
                "themes": None,
                "rating": None,
            },
            {
                "title": "The Midnight Library",
                "author": "Ma",
                "themes": None,
                "rating": None,
            },
            {
                "title": "The Midnight Library",
                "author": "Matt Haig",
                "themes": None,
                "rating": None,
            },
            {
                "title": "The Midnight Library",
                "author": "Matt Haig",
                "themes": None,
                "rating": None,
            },
            {
                "title": "The Midnight Library",
                "author": "Matt Haig",
                "themes": ["Life c"],
                "rating": None,
            },
            {
                "title": "The Midnight Library",
                "author": "Matt Haig",
                "themes": ["Life choi"],
                "rating": None,
            },
            {
                "title": "The Midnight Library",
                "author": "Matt Haig",
                "themes": ["Life choice"],
                "rating": None,
            },
            {
                "title": "The Midnight Library",
                "author": "Matt Haig",
                "themes": ["Life choices", ""],
                "rating": None,
            },
            {
                "title": "The Midnight Library",
                "author": "Matt Haig",
                "themes": ["Life choices", "Second ch"],
                "rating": None,
            },
            {
                "title": "The Midnight Library",
                "author": "Matt Haig",
                "themes": ["Life choices", "Second chances"],
                "rating": None,
            },
            {
                "title": "The Midnight Library",
                "author": "Matt Haig",
                "themes": ["Life choices", "Second chances", "Self-"],
                "rating": None,
            },
            {
                "title": "The Midnight Library",
                "author": "Matt Haig",
                "themes": ["Life choices", "Second chances", "Self-discove"],
                "rating": None,
            },
            {
                "title": "The Midnight Library",
                "author": "Matt Haig",
                "themes": ["Life choices", "Second chances", "Self-discovery", "Re"],
                "rating": None,
            },
            {
                "title": "The Midnight Library",
                "author": "Matt Haig",
                "themes": [
                    "Life choices",
                    "Second chances",
                    "Self-discovery",
                    "Regret and ",
                ],
                "rating": None,
            },
            {
                "title": "The Midnight Library",
                "author": "Matt Haig",
                "themes": [
                    "Life choices",
                    "Second chances",
                    "Self-discovery",
                    "Regret and redemption",
                ],
                "rating": None,
            },
            {
                "title": "The Midnight Library",
                "author": "Matt Haig",
                "themes": [
                    "Life choices",
                    "Second chances",
                    "Self-discovery",
                    "Regret and redemption",
                ],
                "rating": None,
            },
            {
                "title": "The Midnight Library",
                "author": "Matt Haig",
                "themes": [
                    "Life choices",
                    "Second chances",
                    "Self-discovery",
                    "Regret and redemption",
                ],
                "rating": 5,
            },
        ],
    }
)



from inline_snapshot import snapshot

test_snapshot = snapshot(
    {
        "exception": {
            "type": "FeatureNotSupportedError",
            "args": '("Can\'t use completions client for responses model: openai/gpt-4o-mini:responses",)',
            "feature": "responses API",
            "model_id": "openai/gpt-4o-mini:responses",
            "provider_id": "openai:completions",
        }
    }
)



from mirascope import ops


@ops.version
def process_data(data: str) -> str:
    """Process the input data."""
    return f"Processed: {data}"


# Access version info
if (info := process_data.version_info) is not None:
    print(f"Hash: {info.hash}")
    print(f"Version: {info.version}")
    print(f"Name: {info.name}")

# Call the function
result = process_data("example")
print(result)



"""Tests for Span explicit tracing functionality."""

from __future__ import annotations

import logging
from collections.abc import Generator

import pytest
from inline_snapshot import snapshot
from opentelemetry import trace as otel_trace
from opentelemetry.sdk.trace.export.in_memory_span_exporter import InMemorySpanExporter
from opentelemetry.trace import (
    NonRecordingSpan,
    SpanContext,
    TraceFlags,
    TraceState,
    format_span_id,
    format_trace_id,
)

import mirascope

from ..utils import extract_span_data


class _StaticNoOpTracer:
    """A static no-op tracer that always returns non-recording spans."""

    def start_span(self, name: str) -> NonRecordingSpan:
        """Create a new non-recording span."""
        context = SpanContext(
            trace_id=0x1234567890ABCDEF1234567890ABCDEF,
            span_id=0x1234567890ABCDEF,
            is_remote=False,
            trace_flags=TraceFlags(0x01),
            trace_state=TraceState(),
        )
        return NonRecordingSpan(context)


class _StaticNoOpProvider:
    """A static no-op tracer provider for testing noop span behavior."""

    def get_tracer(
        self,
        name: str,
        version: str | None = None,
        schema_url: str | None = None,
        attributes: dict[str, object] | None = None,
    ) -> _StaticNoOpTracer:
        return _StaticNoOpTracer()


@pytest.fixture
def noop_provider() -> Generator[None, None, None]:
    """Fixture that provides a no-op tracer provider for testing."""
    original_provider = otel_trace._TRACER_PROVIDER  # pyright: ignore[reportPrivateUsage]
    otel_trace._TRACER_PROVIDER = _StaticNoOpProvider()  # pyright: ignore[reportPrivateUsage]
    try:
        yield
    finally:
        otel_trace._TRACER_PROVIDER = original_provider  # pyright: ignore[reportPrivateUsage]


def test_noop_span_basic(noop_provider: None) -> None:
    """Test that no-op span handles all operations safely without errors."""
    with mirascope.ops.span("noop") as span:
        assert span.is_noop is True
        assert span.span_id is None

        span.set(a=1, b="x")
        span.event("check", ok=True)
        span.debug("dbg")
        span.info("inf")
        span.warning("warn")
        span.error("err")
        span.critical("crit")


def test_noop_span_with_exception(noop_provider: None) -> None:
    """Test that no-op span handles exceptions gracefully without recording."""
    with (
        pytest.raises(ValueError, match="test error"),
        mirascope.ops.span("error-phase") as span,
    ):
        span.set(test=True)
        raise ValueError("test error")


def test_span_records_attributes_and_events(
    span_exporter: InMemorySpanExporter,
) -> None:
    """Test that configured span records attributes and events correctly."""
    with mirascope.ops.span("op") as span:
        assert span.is_noop is False
        assert span.span_id is not None

        span.set(a=1, b="x")
        span.event("evt", ok=True)
        span.error("boom", code=500)

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 1

    span = spans[0]
    assert extract_span_data(span) == snapshot(
        {
            "name": "op",
            "attributes": {"mirascope.type": "trace", "a": 1, "b": "x"},
            "status": {"status_code": "ERROR", "description": None},
            "events": [
                {"name": "evt", "attributes": {"ok": True}},
                {
                    "name": "error",
                    "attributes": {"message": "boom", "level": "error", "code": 500},
                },
            ],
        }
    )


def test_span_full_lifecycle(span_exporter: InMemorySpanExporter) -> None:
    """Test complete span lifecycle with all logging methods and attributes."""
    with mirascope.ops.span("lifecycle") as span:
        span.set(phase="start", version="1.0", env="test")
        span.debug("Starting process", step=1)
        span.info("Process initialized", ready=True)
        span.warning("Low memory", available_mb=256)
        span.event("custom", data='{"key":"value"}')
        span.error("Connection failed", host="localhost")
        span.critical("System failure", code="CRITICAL_001")

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 1

    span = spans[0]
    assert extract_span_data(span) == snapshot(
        {
            "name": "lifecycle",
            "attributes": {
                "mirascope.type": "trace",
                "phase": "start",
                "version": "1.0",
                "env": "test",
            },
            "status": {"status_code": "ERROR", "description": None},
            "events": [
                {
                    "name": "debug",
                    "attributes": {
                        "message": "Starting process",
                        "level": "debug",
                        "step": 1,
                    },
                },
                {
                    "name": "info",
                    "attributes": {
                        "message": "Process initialized",
                        "level": "info",
                        "ready": True,
                    },
                },
                {
                    "name": "warning",
                    "attributes": {
                        "message": "Low memory",
                        "level": "warning",
                        "available_mb": 256,
                    },
                },
                {"name": "custom", "attributes": {"data": '{"key":"value"}'}},
                {
                    "name": "error",
                    "attributes": {
                        "message": "Connection failed",
                        "level": "error",
                        "host": "localhost",
                    },
                },
                {
                    "name": "critical",
                    "attributes": {
                        "message": "System failure",
                        "level": "error",
                        "code": "CRITICAL_001",
                    },
                },
            ],
        }
    )


def test_span_with_exception_recording(span_exporter: InMemorySpanExporter) -> None:
    """Test that span records exceptions properly with ERROR status."""
    with (
        pytest.raises(RuntimeError, match="Something went wrong"),
        mirascope.ops.span("exception-test") as span,
    ):
        span.set(operation="risky")
        raise RuntimeError("Something went wrong")

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 1

    span = spans[0]
    span_data = extract_span_data(span)
    assert span_data["name"] == "exception-test"
    assert span_data["attributes"] == snapshot(
        {"mirascope.type": "trace", "operation": "risky"}
    )
    assert span_data["status"] == snapshot(
        {"status_code": "ERROR", "description": None}
    )

    events = span_data.get("events", [])
    assert isinstance(events, list)
    assert len([e for e in events if e["name"] == "exception"]) == 1


def test_span_nested_contexts(span_exporter: InMemorySpanExporter) -> None:
    """Test nested span contexts create proper parent-child relationships."""
    with mirascope.ops.span("parent") as parent:
        parent.set(level=1)
        with mirascope.ops.span("child") as child:
            child.set(level=2)
            with mirascope.ops.span("grandchild") as grandchild:
                grandchild.set(level=3)

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 3

    assert extract_span_data(spans[0]) == snapshot(
        {
            "name": "grandchild",
            "attributes": {"mirascope.type": "trace", "level": 3},
            "status": {"status_code": "UNSET", "description": None},
            "events": [],
        }
    )
    assert extract_span_data(spans[1]) == snapshot(
        {
            "name": "child",
            "attributes": {"mirascope.type": "trace", "level": 2},
            "status": {"status_code": "UNSET", "description": None},
            "events": [],
        }
    )
    assert extract_span_data(spans[2]) == snapshot(
        {
            "name": "parent",
            "attributes": {"mirascope.type": "trace", "level": 1},
            "status": {"status_code": "UNSET", "description": None},
            "events": [],
        }
    )


def test_span_id_format(span_exporter: InMemorySpanExporter) -> None:
    """Test that span_id is properly formatted as 16-character hex string."""
    with mirascope.ops.span("format-test") as span:
        span_id = span.span_id
        assert span_id is not None
        assert len(span_id) == 16
        assert all(c in "0123456789abcdef" for c in span_id)


def test_span_concurrent_operations(span_exporter: InMemorySpanExporter) -> None:
    """Test that multiple spans can operate independently without interference."""
    with mirascope.ops.span("operation-1") as span1:
        span1.set(task="first")
        with mirascope.ops.span("operation-2") as span2:
            span2.set(task="second")
            span1.event("still-active")
            span2.event("concurrent")

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 2

    assert extract_span_data(spans[0]) == snapshot(
        {
            "name": "operation-2",
            "attributes": {"mirascope.type": "trace", "task": "second"},
            "status": {"status_code": "UNSET", "description": None},
            "events": [{"name": "concurrent", "attributes": {}}],
        }
    )
    assert extract_span_data(spans[1]) == snapshot(
        {
            "name": "operation-1",
            "attributes": {"mirascope.type": "trace", "task": "first"},
            "status": {"status_code": "UNSET", "description": None},
            "events": [{"name": "still-active", "attributes": {}}],
        }
    )


def test_noop_span_logs_once(
    caplog: pytest.LogCaptureFixture, noop_provider: _StaticNoOpProvider
) -> None:
    """Test that warning logs exactly once when spans operate in no-op mode."""
    from mirascope.ops._internal import spans as span_module

    span_module._warned_noop = False  # pyright: ignore[reportPrivateUsage]
    with caplog.at_level(logging.WARNING, logger="mirascope"):
        with mirascope.ops.span("noop-log") as first:
            assert first.is_noop is True
            assert first.span_id is None
            assert first.trace_id is None
            assert first.span_context is None
        with mirascope.ops.span("noop-log-second") as second:
            assert second.is_noop is True
            assert second.span_id is None
            assert second.trace_id is None
            assert second.span_context is None

    warnings = [
        record
        for record in caplog.records
        if "mirascope tracing is not configured" in record.getMessage()
    ]
    assert len(warnings) == 1


def test_span_preserves_complex_attributes(
    span_exporter: InMemorySpanExporter,
) -> None:
    """Test that complex attributes and events are preserved as sequences/strings."""
    with mirascope.ops.span("complex") as span:
        span.set(numbers=[1, 2, 3], payload='{"ok": true}')
        span.event("details", points=[1, 2], info='{"nested": "yes"}')

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 1

    data = extract_span_data(spans[0])
    assert data == snapshot(
        {
            "name": "complex",
            "attributes": {
                "mirascope.type": "trace",
                "numbers": (1, 2, 3),
                "payload": '{"ok": true}',
            },
            "status": {"status_code": "UNSET", "description": None},
            "events": [
                {
                    "name": "details",
                    "attributes": {
                        "points": (1, 2),
                        "info": '{"nested": "yes"}',
                    },
                }
            ],
        }
    )


def test_span_trace_and_context_properties(span_exporter: InMemorySpanExporter) -> None:
    """Test that trace/span IDs are exposed and context round-trips correctly."""
    with mirascope.ops.span("context-check") as span:
        trace_id = span.trace_id
        span_id = span.span_id
        context = span.span_context

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 1

    otel_context = spans[0].get_span_context()
    assert otel_context is not None
    assert trace_id == format_trace_id(otel_context.trace_id)
    assert span_id == format_span_id(otel_context.span_id)
    assert context == otel_context


def test_span_with_session(span_exporter: InMemorySpanExporter) -> None:
    """Test that span records session ID from active session."""
    with (
        mirascope.ops.session(id="session-123"),
        mirascope.ops.span("op-with-session") as span,
    ):
        span.set(task="test")

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 1

    span_data = extract_span_data(spans[0])
    assert span_data == snapshot(
        {
            "name": "op-with-session",
            "attributes": {
                "mirascope.type": "trace",
                "mirascope.ops.session.id": "session-123",
                "task": "test",
            },
            "status": {"status_code": "UNSET", "description": None},
            "events": [],
        }
    )


def test_span_with_session_attributes(span_exporter: InMemorySpanExporter) -> None:
    """Test that span records session ID and attributes."""
    with (
        mirascope.ops.session(
            id="session-456", attributes={"user": "alice", "env": "prod"}
        ),
        mirascope.ops.span("op-with-attrs"),
    ):
        pass

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 1

    span_data = extract_span_data(spans[0])
    assert span_data == snapshot(
        {
            "name": "op-with-attrs",
            "attributes": {
                "mirascope.type": "trace",
                "mirascope.ops.session.id": "session-456",
                "mirascope.ops.session.attributes": '{"user":"alice","env":"prod"}',
            },
            "status": {"status_code": "UNSET", "description": None},
            "events": [],
        }
    )


def test_span_without_session(span_exporter: InMemorySpanExporter) -> None:
    """Test that span works without active session."""
    with mirascope.ops.span("op-no-session") as span:
        span.set(task="test")

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 1

    span_data = extract_span_data(spans[0])
    assert span_data == snapshot(
        {
            "name": "op-no-session",
            "attributes": {"mirascope.type": "trace", "task": "test"},
            "status": {"status_code": "UNSET", "description": None},
            "events": [],
        }
    )


def test_span_nested_session_override(span_exporter: InMemorySpanExporter) -> None:
    """Test that nested session correctly overrides parent session."""
    with mirascope.ops.session(id="outer-session"):
        with mirascope.ops.span("outer-span"):
            pass

        with (
            mirascope.ops.session(id="inner-session"),
            mirascope.ops.span("inner-span"),
        ):
            pass

        with mirascope.ops.span("outer-span-2"):
            pass

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 3

    assert extract_span_data(spans[0]) == snapshot(
        {
            "name": "outer-span",
            "attributes": {
                "mirascope.type": "trace",
                "mirascope.ops.session.id": "outer-session",
            },
            "status": {"status_code": "UNSET", "description": None},
            "events": [],
        }
    )
    assert extract_span_data(spans[1]) == snapshot(
        {
            "name": "inner-span",
            "attributes": {
                "mirascope.type": "trace",
                "mirascope.ops.session.id": "inner-session",
            },
            "status": {"status_code": "UNSET", "description": None},
            "events": [],
        }
    )
    assert extract_span_data(spans[2]) == snapshot(
        {
            "name": "outer-span-2",
            "attributes": {
                "mirascope.type": "trace",
                "mirascope.ops.session.id": "outer-session",
            },
            "status": {"status_code": "UNSET", "description": None},
            "events": [],
        }
    )


def test_span_with_initial_attributes(span_exporter: InMemorySpanExporter) -> None:
    """Test that span accepts initial attributes in constructor."""
    with mirascope.ops.span("init-attrs", user_id="123", action="process") as span:
        span.set(additional="value")

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 1

    span_data = extract_span_data(spans[0])
    assert span_data == snapshot(
        {
            "name": "init-attrs",
            "attributes": {
                "mirascope.type": "trace",
                "user_id": "123",
                "action": "process",
                "additional": "value",
            },
            "status": {"status_code": "UNSET", "description": None},
            "events": [],
        }
    )


def test_span_with_initial_complex_attributes(
    span_exporter: InMemorySpanExporter,
) -> None:
    """Test that span accepts complex initial attributes without serialization."""
    with mirascope.ops.span("complex-init", tags=["a", "b"], config='{"key":"val"}'):
        pass

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 1

    span_data = extract_span_data(spans[0])
    assert span_data == snapshot(
        {
            "name": "complex-init",
            "attributes": {
                "mirascope.type": "trace",
                "tags": ("a", "b"),
                "config": '{"key":"val"}',
            },
            "status": {"status_code": "UNSET", "description": None},
            "events": [],
        }
    )


def test_noop_span_with_initial_attributes(noop_provider: None) -> None:
    """Test that no-op span handles initial attributes without errors."""
    with mirascope.ops.span("noop-init", user_id="456", count=3) as span:
        assert span.is_noop is True
        assert span.span_id is None



# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
import typing_extensions
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from ..core.serialization import FieldMetadata
from .rate_limit_error_tag import RateLimitErrorTag


class RateLimitError(UniversalBaseModel):
    message: str
    organization_id: typing_extensions.Annotated[
        str, FieldMetadata(alias="organizationId")
    ]
    limit: float
    retry_after: typing_extensions.Annotated[float, FieldMetadata(alias="retryAfter")]
    plan_tier: typing_extensions.Annotated[str, FieldMetadata(alias="planTier")]
    tag: RateLimitErrorTag

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(
            extra="allow", frozen=True
        )  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



from pydantic import BaseModel, field_validator

from mirascope import llm


class BookRecommendation(BaseModel):
    title: str
    year: int

    @field_validator("year")
    @classmethod
    def must_be_recent(cls, v: int) -> int:
        if v < 2020:
            raise ValueError(f"Must be published after 2020, got {v}")
        return v


@llm.call("openai/gpt-4o-mini", format=BookRecommendation)
def recommend_book():
    return "Recommend a science fiction book."


response = recommend_book()
book, response = response.validate()  # Retries once if validation fails
print(f"{book.title} ({book.year})")



# This file was auto-generated by Fern from our API Definition.

import typing

from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.request_options import RequestOptions
from .raw_client import (
    AsyncRawOrganizationInvitationsClient,
    RawOrganizationInvitationsClient,
)
from .types.organization_invitations_accept_response import (
    OrganizationInvitationsAcceptResponse,
)
from .types.organization_invitations_create_request_role import (
    OrganizationInvitationsCreateRequestRole,
)
from .types.organization_invitations_create_response import (
    OrganizationInvitationsCreateResponse,
)
from .types.organization_invitations_get_response import (
    OrganizationInvitationsGetResponse,
)
from .types.organization_invitations_list_response_item import (
    OrganizationInvitationsListResponseItem,
)

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class OrganizationInvitationsClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._raw_client = RawOrganizationInvitationsClient(
            client_wrapper=client_wrapper
        )

    @property
    def with_raw_response(self) -> RawOrganizationInvitationsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        RawOrganizationInvitationsClient
        """
        return self._raw_client

    def list(
        self,
        organization_id: str,
        *,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> typing.List[OrganizationInvitationsListResponseItem]:
        """
        Parameters
        ----------
        organization_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        typing.List[OrganizationInvitationsListResponseItem]
            Success

        Examples
        --------
        from mirascope.api._generated import Mirascope

        client = Mirascope()
        client.organization_invitations.list(
            organization_id="organizationId",
        )
        """
        _response = self._raw_client.list(
            organization_id, request_options=request_options
        )
        return _response.data

    def create(
        self,
        organization_id: str,
        *,
        recipient_email: str,
        role: OrganizationInvitationsCreateRequestRole,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> OrganizationInvitationsCreateResponse:
        """
        Parameters
        ----------
        organization_id : str

        recipient_email : str
            a string matching the pattern ^[^ \t\n\r\f\v@]+@[^ \t\n\r\f\v@]+[.][^ \t\n\r\f\v@]+$

        role : OrganizationInvitationsCreateRequestRole

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        OrganizationInvitationsCreateResponse
            Success

        Examples
        --------
        from mirascope.api._generated import Mirascope

        client = Mirascope()
        client.organization_invitations.create(
            organization_id="organizationId",
            recipient_email="recipientEmail",
            role="ADMIN",
        )
        """
        _response = self._raw_client.create(
            organization_id,
            recipient_email=recipient_email,
            role=role,
            request_options=request_options,
        )
        return _response.data

    def get(
        self,
        organization_id: str,
        invitation_id: str,
        *,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> OrganizationInvitationsGetResponse:
        """
        Parameters
        ----------
        organization_id : str

        invitation_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        OrganizationInvitationsGetResponse
            Success

        Examples
        --------
        from mirascope.api._generated import Mirascope

        client = Mirascope()
        client.organization_invitations.get(
            organization_id="organizationId",
            invitation_id="invitationId",
        )
        """
        _response = self._raw_client.get(
            organization_id, invitation_id, request_options=request_options
        )
        return _response.data

    def resend(
        self,
        organization_id: str,
        invitation_id: str,
        *,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Parameters
        ----------
        organization_id : str

        invitation_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from mirascope.api._generated import Mirascope

        client = Mirascope()
        client.organization_invitations.resend(
            organization_id="organizationId",
            invitation_id="invitationId",
        )
        """
        _response = self._raw_client.resend(
            organization_id, invitation_id, request_options=request_options
        )
        return _response.data

    def revoke(
        self,
        organization_id: str,
        invitation_id: str,
        *,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Parameters
        ----------
        organization_id : str

        invitation_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from mirascope.api._generated import Mirascope

        client = Mirascope()
        client.organization_invitations.revoke(
            organization_id="organizationId",
            invitation_id="invitationId",
        )
        """
        _response = self._raw_client.revoke(
            organization_id, invitation_id, request_options=request_options
        )
        return _response.data

    def accept(
        self, *, token: str, request_options: typing.Optional[RequestOptions] = None
    ) -> OrganizationInvitationsAcceptResponse:
        """
        Parameters
        ----------
        token : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        OrganizationInvitationsAcceptResponse
            Success

        Examples
        --------
        from mirascope.api._generated import Mirascope

        client = Mirascope()
        client.organization_invitations.accept(
            token="token",
        )
        """
        _response = self._raw_client.accept(
            token=token, request_options=request_options
        )
        return _response.data


class AsyncOrganizationInvitationsClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._raw_client = AsyncRawOrganizationInvitationsClient(
            client_wrapper=client_wrapper
        )

    @property
    def with_raw_response(self) -> AsyncRawOrganizationInvitationsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        AsyncRawOrganizationInvitationsClient
        """
        return self._raw_client

    async def list(
        self,
        organization_id: str,
        *,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> typing.List[OrganizationInvitationsListResponseItem]:
        """
        Parameters
        ----------
        organization_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        typing.List[OrganizationInvitationsListResponseItem]
            Success

        Examples
        --------
        import asyncio

        from mirascope.api._generated import AsyncMirascope

        client = AsyncMirascope()


        async def main() -> None:
            await client.organization_invitations.list(
                organization_id="organizationId",
            )


        asyncio.run(main())
        """
        _response = await self._raw_client.list(
            organization_id, request_options=request_options
        )
        return _response.data

    async def create(
        self,
        organization_id: str,
        *,
        recipient_email: str,
        role: OrganizationInvitationsCreateRequestRole,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> OrganizationInvitationsCreateResponse:
        """
        Parameters
        ----------
        organization_id : str

        recipient_email : str
            a string matching the pattern ^[^ \t\n\r\f\v@]+@[^ \t\n\r\f\v@]+[.][^ \t\n\r\f\v@]+$

        role : OrganizationInvitationsCreateRequestRole

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        OrganizationInvitationsCreateResponse
            Success

        Examples
        --------
        import asyncio

        from mirascope.api._generated import AsyncMirascope

        client = AsyncMirascope()


        async def main() -> None:
            await client.organization_invitations.create(
                organization_id="organizationId",
                recipient_email="recipientEmail",
                role="ADMIN",
            )


        asyncio.run(main())
        """
        _response = await self._raw_client.create(
            organization_id,
            recipient_email=recipient_email,
            role=role,
            request_options=request_options,
        )
        return _response.data

    async def get(
        self,
        organization_id: str,
        invitation_id: str,
        *,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> OrganizationInvitationsGetResponse:
        """
        Parameters
        ----------
        organization_id : str

        invitation_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        OrganizationInvitationsGetResponse
            Success

        Examples
        --------
        import asyncio

        from mirascope.api._generated import AsyncMirascope

        client = AsyncMirascope()


        async def main() -> None:
            await client.organization_invitations.get(
                organization_id="organizationId",
                invitation_id="invitationId",
            )


        asyncio.run(main())
        """
        _response = await self._raw_client.get(
            organization_id, invitation_id, request_options=request_options
        )
        return _response.data

    async def resend(
        self,
        organization_id: str,
        invitation_id: str,
        *,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Parameters
        ----------
        organization_id : str

        invitation_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        import asyncio

        from mirascope.api._generated import AsyncMirascope

        client = AsyncMirascope()


        async def main() -> None:
            await client.organization_invitations.resend(
                organization_id="organizationId",
                invitation_id="invitationId",
            )


        asyncio.run(main())
        """
        _response = await self._raw_client.resend(
            organization_id, invitation_id, request_options=request_options
        )
        return _response.data

    async def revoke(
        self,
        organization_id: str,
        invitation_id: str,
        *,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Parameters
        ----------
        organization_id : str

        invitation_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        import asyncio

        from mirascope.api._generated import AsyncMirascope

        client = AsyncMirascope()


        async def main() -> None:
            await client.organization_invitations.revoke(
                organization_id="organizationId",
                invitation_id="invitationId",
            )


        asyncio.run(main())
        """
        _response = await self._raw_client.revoke(
            organization_id, invitation_id, request_options=request_options
        )
        return _response.data

    async def accept(
        self, *, token: str, request_options: typing.Optional[RequestOptions] = None
    ) -> OrganizationInvitationsAcceptResponse:
        """
        Parameters
        ----------
        token : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        OrganizationInvitationsAcceptResponse
            Success

        Examples
        --------
        import asyncio

        from mirascope.api._generated import AsyncMirascope

        client = AsyncMirascope()


        async def main() -> None:
            await client.organization_invitations.accept(
                token="token",
            )


        asyncio.run(main())
        """
        _response = await self._raw_client.accept(
            token=token, request_options=request_options
        )
        return _response.data



# This file was auto-generated by Fern from our API Definition.

# isort: skip_file

from .types import (
    OrganizationMembershipsListResponseItem,
    OrganizationMembershipsListResponseItemRole,
    OrganizationMembershipsUpdateRequestRole,
    OrganizationMembershipsUpdateResponse,
    OrganizationMembershipsUpdateResponseRole,
)

__all__ = [
    "OrganizationMembershipsListResponseItem",
    "OrganizationMembershipsListResponseItemRole",
    "OrganizationMembershipsUpdateRequestRole",
    "OrganizationMembershipsUpdateResponse",
    "OrganizationMembershipsUpdateResponseRole",
]



"""Retry stream response wrappers that add retry capabilities to streaming LLM responses."""

from collections.abc import AsyncIterator, Awaitable, Callable, Iterator
from typing import TYPE_CHECKING, Generic, overload

from ..content import (
    AssistantContentChunk,
)
from ..context import Context, DepsT
from ..exceptions import RetriesExhausted, StreamRestarted
from ..formatting import FormattableT
from ..messages import UserContent
from ..models import Model
from ..responses import (
    AsyncContextStreamResponse,
    AsyncStreamResponse,
    ContextStreamResponse,
    StreamResponse,
)
from .utils import RetryFailure, get_retry_model_from_context

if TYPE_CHECKING:
    from .retry_config import RetryConfig
    from .retry_models import RetryModel


class RetryStreamResponse(StreamResponse[FormattableT]):
    """A streaming response wrapper that includes retry capabilities.

    Extends `StreamResponse` directly, copying all attributes from a wrapped response
    and adding retry configuration.

    This wraps a `StreamResponse` and adds automatic retry behavior when
    retryable errors occur during iteration. When a retry happens, a
    `StreamRestarted` exception is raised so the user can handle the restart
    (e.g., clear terminal output) before re-iterating.

    Supports fallback models - when the active model exhausts its retries,
    the next fallback model is tried.

    Example:
        ```python
        response = retry_model.stream("Tell me a story")

        while True:
            try:
                for chunk in response.text_stream():
                    print(chunk, end="", flush=True)
                break  # Success
            except llm.StreamRestarted as e:
                print(e.message)
        ```
    """

    _current_variant: "RetryModel"
    """The current RetryModel variant being used."""

    _variants_iter: Iterator["RetryModel"]
    """Iterator over model variants with backoff delays."""

    retry_failures: list[RetryFailure]
    """Failed attempts before success (empty if first attempt succeeded)."""

    _stream_fn: Callable[[Model], StreamResponse[FormattableT]]

    def __init__(
        self,
        retry_model: "RetryModel",
        stream_fn: Callable[[Model], StreamResponse[FormattableT]],
    ) -> None:
        """Initialize a RetryStreamResponse.

        Args:
            retry_model: The RetryModel providing retry configuration.
            stream_fn: Function that creates a stream from a Model.
        """
        self._variants_iter = retry_model.variants()
        self._current_variant = next(self._variants_iter)
        self._stream_fn = stream_fn
        self.retry_failures = []

        # Create the initial stream and copy all attributes
        initial_stream = stream_fn(self._current_variant.get_active_model())
        for key, value in initial_stream.__dict__.items():
            object.__setattr__(self, key, value)

    @property
    def model(self) -> "RetryModel":
        """A RetryModel with parameters matching this response.

        If a model is set in context (via `llm.model()` or `llm.retry_model()`),
        that model is used instead, wrapped with this response's retry config.
        """
        return get_retry_model_from_context(self._current_variant)

    @property
    def retry_config(self) -> "RetryConfig":
        """The retry configuration for this response."""
        return self.model.retry_config

    def _reset_stream(self, variant: "RetryModel") -> None:
        """Reset to a fresh stream for a new retry attempt."""
        new_stream = self._stream_fn(variant.get_active_model())
        # Copy all attributes from the new stream
        for key, value in new_stream.__dict__.items():
            object.__setattr__(self, key, value)

    def chunk_stream(self) -> Iterator[AssistantContentChunk]:
        """Returns an iterator that yields content chunks with retry support.

        If a retryable error occurs during iteration, the stream is reset
        and a `StreamRestarted` exception is raised. The user should catch
        this exception and re-iterate to continue from the new attempt.

        Raises:
            StreamRestarted: When the stream is reset for a retry attempt.
            Exception: The underlying error if max retries are exhausted.
        """
        config = self._current_variant.retry_config
        try:
            yield from super().chunk_stream()
        except config.retry_on as e:
            failure = RetryFailure(
                model=self._current_variant.get_active_model(), exception=e
            )
            self.retry_failures.append(failure)

            # Try to get next variant (handles backoff and fallback)
            try:
                self._current_variant = next(self._variants_iter)
            except StopIteration:
                raise RetriesExhausted(self.retry_failures) from None

            self._reset_stream(self._current_variant)

            raise StreamRestarted(
                failure=failure,
            ) from e

    @overload
    def resume(
        self: "RetryStreamResponse[None]", content: UserContent
    ) -> "RetryStreamResponse[None]": ...

    @overload
    def resume(
        self: "RetryStreamResponse[FormattableT]", content: UserContent
    ) -> "RetryStreamResponse[FormattableT]": ...

    def resume(
        self, content: UserContent
    ) -> "RetryStreamResponse[None] | RetryStreamResponse[FormattableT]":
        """Generate a new RetryStreamResponse using this response's messages with additional user content."""
        return self.model.resume_stream(response=self, content=content)


class AsyncRetryStreamResponse(AsyncStreamResponse[FormattableT]):
    """An async streaming response wrapper that includes retry capabilities.

    Extends `AsyncStreamResponse` directly, copying all attributes from a wrapped response
    and adding retry configuration.

    This wraps an `AsyncStreamResponse` and adds automatic retry behavior when
    retryable errors occur during iteration. When a retry happens, a
    `StreamRestarted` exception is raised so the user can handle the restart
    (e.g., clear terminal output) before re-iterating.

    Supports fallback models - when the active model exhausts its retries,
    the next fallback model is tried.

    Example:
        ```python
        response = await retry_model.stream_async("Tell me a story")

        while True:
            try:
                async for chunk in response.text_stream():
                    print(chunk, end="", flush=True)
                break  # Success
            except llm.StreamRestarted as e:
                print(e.message)
        ```
    """

    _current_variant: "RetryModel"
    """The current RetryModel variant being used."""

    _variants_iter: AsyncIterator["RetryModel"]
    """Async iterator over model variants with backoff delays."""

    retry_failures: list[RetryFailure]
    """Failed attempts before success (empty if first attempt succeeded)."""

    _stream_fn: Callable[[Model], Awaitable[AsyncStreamResponse[FormattableT]]]

    def __init__(
        self,
        stream_fn: Callable[[Model], Awaitable[AsyncStreamResponse[FormattableT]]],
        initial_stream: AsyncStreamResponse[FormattableT],
        initial_variant: "RetryModel",
        variants_iter: AsyncIterator["RetryModel"],
    ) -> None:
        """Initialize an AsyncRetryStreamResponse.

        Args:
            stream_fn: Async function that creates a stream from a Model.
            initial_stream: The pre-awaited initial stream.
            initial_variant: The first variant from the iterator.
            variants_iter: The async iterator for remaining variants.
        """
        # Copy all attributes from the initial stream
        for key, value in initial_stream.__dict__.items():
            object.__setattr__(self, key, value)

        self._current_variant = initial_variant
        self._variants_iter = variants_iter
        self._stream_fn = stream_fn
        self.retry_failures = []

    @property
    def model(self) -> "RetryModel":
        """A RetryModel with parameters matching this response.

        If a model is set in context (via `llm.model()` or `llm.retry_model()`),
        that model is used instead, wrapped with this response's retry config.
        """
        return get_retry_model_from_context(self._current_variant)

    @property
    def retry_config(self) -> "RetryConfig":
        """The retry configuration for this response."""
        return self.model.retry_config

    async def _reset_stream(self, variant: "RetryModel") -> None:
        """Reset to a fresh stream for a new retry attempt."""
        new_stream = await self._stream_fn(variant.get_active_model())
        # Copy all attributes from the new stream
        for key, value in new_stream.__dict__.items():
            object.__setattr__(self, key, value)

    async def chunk_stream(self) -> AsyncIterator[AssistantContentChunk]:
        """Returns an async iterator that yields content chunks with retry support.

        If a retryable error occurs during iteration, the stream is reset
        and a `StreamRestarted` exception is raised. The user should catch
        this exception and re-iterate to continue from the new attempt.

        Raises:
            StreamRestarted: When the stream is reset for a retry attempt.
            Exception: The underlying error if max retries are exhausted.
        """
        config = self._current_variant.retry_config
        try:
            async for chunk in super().chunk_stream():
                yield chunk
        except config.retry_on as e:
            failure = RetryFailure(
                model=self._current_variant.get_active_model(), exception=e
            )
            self.retry_failures.append(failure)

            # Try to get next variant (handles backoff and fallback)
            try:
                self._current_variant = await anext(self._variants_iter)
            except StopAsyncIteration:
                raise RetriesExhausted(self.retry_failures) from None

            await self._reset_stream(self._current_variant)

            raise StreamRestarted(
                failure=failure,
            ) from e

    @overload
    async def resume(
        self: "AsyncRetryStreamResponse[None]", content: UserContent
    ) -> "AsyncRetryStreamResponse[None]": ...

    @overload
    async def resume(
        self: "AsyncRetryStreamResponse[FormattableT]", content: UserContent
    ) -> "AsyncRetryStreamResponse[FormattableT]": ...

    async def resume(
        self, content: UserContent
    ) -> "AsyncRetryStreamResponse[None] | AsyncRetryStreamResponse[FormattableT]":
        """Generate a new AsyncRetryStreamResponse using this response's messages with additional user content."""
        return await self.model.resume_stream_async(response=self, content=content)


class ContextRetryStreamResponse(
    ContextStreamResponse[DepsT, FormattableT], Generic[DepsT, FormattableT]
):
    """A context-aware streaming response wrapper that includes retry capabilities.

    Extends `ContextStreamResponse` directly, copying all attributes from a wrapped response
    and adding retry configuration.

    This wraps a `ContextStreamResponse` and adds automatic retry behavior when
    retryable errors occur during iteration. When a retry happens, a
    `StreamRestarted` exception is raised so the user can handle the restart
    (e.g., clear terminal output) before re-iterating.

    Supports fallback models - when the active model exhausts its retries,
    the next fallback model is tried.

    Example:
        ```python
        response = retry_model.context_stream("Tell me a story", ctx=ctx)

        while True:
            try:
                for chunk in response.text_stream():
                    print(chunk, end="", flush=True)
                break  # Success
            except llm.StreamRestarted as e:
                print(e.message)
        ```
    """

    _current_variant: "RetryModel"
    """The current RetryModel variant being used."""

    _variants_iter: Iterator["RetryModel"]
    """Iterator over model variants with backoff delays."""

    retry_failures: list[RetryFailure]
    """Failed attempts before success (empty if first attempt succeeded)."""

    _stream_fn: Callable[[Model], ContextStreamResponse[DepsT, FormattableT]]

    def __init__(
        self,
        retry_model: "RetryModel",
        stream_fn: Callable[[Model], ContextStreamResponse[DepsT, FormattableT]],
    ) -> None:
        """Initialize a ContextRetryStreamResponse.

        Args:
            retry_model: The RetryModel providing retry configuration.
            stream_fn: Function that creates a stream from a Model.
        """
        self._variants_iter = retry_model.variants()
        self._current_variant = next(self._variants_iter)
        self._stream_fn = stream_fn
        self.retry_failures = []

        # Create the initial stream and copy all attributes
        initial_stream = stream_fn(self._current_variant.get_active_model())
        for key, value in initial_stream.__dict__.items():
            object.__setattr__(self, key, value)

    @property
    def model(self) -> "RetryModel":
        """A RetryModel with parameters matching this response.

        If a model is set in context (via `llm.model()` or `llm.retry_model()`),
        that model is used instead, wrapped with this response's retry config.
        """
        return get_retry_model_from_context(self._current_variant)

    @property
    def retry_config(self) -> "RetryConfig":
        """The retry configuration for this response."""
        return self.model.retry_config

    def _reset_stream(self, variant: "RetryModel") -> None:
        """Reset to a fresh stream for a new retry attempt."""
        new_stream = self._stream_fn(variant.get_active_model())
        # Copy all attributes from the new stream
        for key, value in new_stream.__dict__.items():
            object.__setattr__(self, key, value)

    def chunk_stream(self) -> Iterator[AssistantContentChunk]:
        """Returns an iterator that yields content chunks with retry support.

        If a retryable error occurs during iteration, the stream is reset
        and a `StreamRestarted` exception is raised. The user should catch
        this exception and re-iterate to continue from the new attempt.

        Raises:
            StreamRestarted: When the stream is reset for a retry attempt.
            Exception: The underlying error if max retries are exhausted.
        """
        config = self._current_variant.retry_config
        try:
            yield from super().chunk_stream()
        except config.retry_on as e:
            failure = RetryFailure(
                model=self._current_variant.get_active_model(), exception=e
            )
            self.retry_failures.append(failure)

            # Try to get next variant (handles backoff and fallback)
            try:
                self._current_variant = next(self._variants_iter)
            except StopIteration:
                raise RetriesExhausted(self.retry_failures) from None

            self._reset_stream(self._current_variant)

            raise StreamRestarted(
                failure=failure,
            ) from e

    @overload
    def resume(
        self: "ContextRetryStreamResponse[DepsT, None]",
        ctx: Context[DepsT],
        content: UserContent,
    ) -> "ContextRetryStreamResponse[DepsT, None]": ...

    @overload
    def resume(
        self: "ContextRetryStreamResponse[DepsT, FormattableT]",
        ctx: Context[DepsT],
        content: UserContent,
    ) -> "ContextRetryStreamResponse[DepsT, FormattableT]": ...

    def resume(
        self, ctx: Context[DepsT], content: UserContent
    ) -> "ContextRetryStreamResponse[DepsT, None] | ContextRetryStreamResponse[DepsT, FormattableT]":
        """Generate a new ContextRetryStreamResponse using this response's messages with additional user content."""
        return self.model.context_resume_stream(ctx=ctx, response=self, content=content)


class AsyncContextRetryStreamResponse(
    AsyncContextStreamResponse[DepsT, FormattableT], Generic[DepsT, FormattableT]
):
    """An async context-aware streaming response wrapper that includes retry capabilities.

    Extends `AsyncContextStreamResponse` directly, copying all attributes from a wrapped response
    and adding retry configuration.

    This wraps an `AsyncContextStreamResponse` and adds automatic retry behavior when
    retryable errors occur during iteration. When a retry happens, a
    `StreamRestarted` exception is raised so the user can handle the restart
    (e.g., clear terminal output) before re-iterating.

    Supports fallback models - when the active model exhausts its retries,
    the next fallback model is tried.

    Example:
        ```python
        ctx = llm.Context(deps=my_deps)
        response = await retry_model.context_stream_async("Tell me a story", ctx=ctx)

        while True:
            try:
                async for chunk in response.text_stream():
                    print(chunk, end="", flush=True)
                break  # Success
            except llm.StreamRestarted as e:
                print(e.message)
        ```
    """

    _current_variant: "RetryModel"
    """The current RetryModel variant being used."""

    _variants_iter: AsyncIterator["RetryModel"]
    """Async iterator over model variants with backoff delays."""

    retry_failures: list[RetryFailure]
    """Failed attempts before success (empty if first attempt succeeded)."""

    _stream_fn: Callable[
        [Model], Awaitable[AsyncContextStreamResponse[DepsT, FormattableT]]
    ]

    def __init__(
        self,
        stream_fn: Callable[
            [Model], Awaitable[AsyncContextStreamResponse[DepsT, FormattableT]]
        ],
        initial_stream: AsyncContextStreamResponse[DepsT, FormattableT],
        initial_variant: "RetryModel",
        variants_iter: AsyncIterator["RetryModel"],
    ) -> None:
        """Initialize an AsyncContextRetryStreamResponse.

        Args:
            stream_fn: Async function that creates a stream from a Model.
            initial_stream: The pre-awaited initial stream.
            initial_variant: The first variant from the iterator.
            variants_iter: The async iterator for remaining variants.
        """
        # Copy all attributes from the initial stream
        for key, value in initial_stream.__dict__.items():
            object.__setattr__(self, key, value)

        self._current_variant = initial_variant
        self._variants_iter = variants_iter
        self._stream_fn = stream_fn
        self.retry_failures = []

    @property
    def model(self) -> "RetryModel":
        """A RetryModel with parameters matching this response.

        If a model is set in context (via `llm.model()` or `llm.retry_model()`),
        that model is used instead, wrapped with this response's retry config.
        """
        return get_retry_model_from_context(self._current_variant)

    @property
    def retry_config(self) -> "RetryConfig":
        """The retry configuration for this response."""
        return self.model.retry_config

    async def _reset_stream(self, variant: "RetryModel") -> None:
        """Reset to a fresh stream for a new retry attempt."""
        new_stream = await self._stream_fn(variant.get_active_model())
        # Copy all attributes from the new stream
        for key, value in new_stream.__dict__.items():
            object.__setattr__(self, key, value)

    async def chunk_stream(self) -> AsyncIterator[AssistantContentChunk]:
        """Returns an async iterator that yields content chunks with retry support.

        If a retryable error occurs during iteration, the stream is reset
        and a `StreamRestarted` exception is raised. The user should catch
        this exception and re-iterate to continue from the new attempt.

        Raises:
            StreamRestarted: When the stream is reset for a retry attempt.
            Exception: The underlying error if max retries are exhausted.
        """
        config = self._current_variant.retry_config
        try:
            async for chunk in super().chunk_stream():
                yield chunk
        except config.retry_on as e:
            failure = RetryFailure(
                model=self._current_variant.get_active_model(), exception=e
            )
            self.retry_failures.append(failure)

            # Try to get next variant (handles backoff and fallback)
            try:
                self._current_variant = await anext(self._variants_iter)
            except StopAsyncIteration:
                raise RetriesExhausted(self.retry_failures) from None

            await self._reset_stream(self._current_variant)

            raise StreamRestarted(
                failure=failure,
            ) from e

    @overload
    async def resume(
        self: "AsyncContextRetryStreamResponse[DepsT, None]",
        ctx: Context[DepsT],
        content: UserContent,
    ) -> "AsyncContextRetryStreamResponse[DepsT, None]": ...

    @overload
    async def resume(
        self: "AsyncContextRetryStreamResponse[DepsT, FormattableT]",
        ctx: Context[DepsT],
        content: UserContent,
    ) -> "AsyncContextRetryStreamResponse[DepsT, FormattableT]": ...

    async def resume(
        self, ctx: Context[DepsT], content: UserContent
    ) -> "AsyncContextRetryStreamResponse[DepsT, None] | AsyncContextRetryStreamResponse[DepsT, FormattableT]":
        """Generate a new AsyncContextRetryStreamResponse using this response's messages with additional user content."""
        return await self.model.context_resume_stream_async(
            ctx=ctx, response=self, content=content
        )



"""The Mirascope Cloud API client."""

from .._stubs import stub_module_if_missing

# Stub modules for missing optional dependencies BEFORE importing
# This must happen before any imports from these modules
stub_module_if_missing("mirascope.api", "api")

# Now imports work regardless of which packages are installed
# ruff: noqa: E402
from .client import AsyncMirascope, Mirascope
from .settings import get_settings, settings

__all__ = ["AsyncMirascope", "Mirascope", "get_settings", "settings"]



from mirascope import llm


@llm.call("google/gemini-3-pro", thinking={"level": "high", "include_thoughts": True})
def solve(problem: str) -> str:
    return problem


response = solve("What is the first prime number that contains 42 as a substring?")
print(response.pretty())



from mirascope import llm

model = llm.model("openai/gpt-5-mini")
response: llm.Response = model.call("What is the capital of France?")

print(response.text())  # Prints the textual content of the response



from mirascope import llm


@llm.call("openai/gpt-5-mini")
def recommend_book(genre: str):
    return f"Recommend a {genre} book."


response = recommend_book.stream("fantasy")

# Stream and display content
for chunk in response.text_stream():
    print(chunk, end="", flush=True)

# After streaming, access accumulated content like a regular Response
print(f"\nTexts: {len(response.texts)}")
print(f"Content parts: {len(response.content)}")
print(f"Messages in history: {len(response.messages)}")
print(f"Usage: {response.usage}")



from mirascope import llm

model = llm.model("anthropic/claude-sonnet-4-5", max_tokens=40)
response = model.call("Write a long story about a bear.")

# finish_reason is None when the response completes normally
# It's set when the response was cut off or stopped abnormally
if response.finish_reason == llm.FinishReason.MAX_TOKENS:
    print("Response was truncated due to token limit")
elif response.finish_reason is None:
    print("Response completed normally")



"""Shared test helpers for exporter tests."""

import time
from typing import Any
from unittest.mock import Mock

from opentelemetry.sdk.trace import ReadableSpan
from opentelemetry.trace import SpanContext, SpanKind, Status, StatusCode, TraceFlags
from opentelemetry.trace.span import TraceState


def create_mock_span(
    name: str = "test-span",
    span_id: int = 0x1234567890ABCDEF,
    trace_id: int = 0x123456789ABCDEF0123456789ABCDEF0,
    parent_span_id: int | None = None,
    shared_resource: Mock | None = None,
    kind: SpanKind = SpanKind.INTERNAL,
    status_code: StatusCode = StatusCode.OK,
    attributes: dict[str, Any] | None = None,
) -> Mock:
    """Create a mock ReadableSpan with common setup.

    Args:
        name: The span name
        span_id: The span ID (64-bit hex)
        trace_id: The trace ID (128-bit hex)
        parent_span_id: Optional parent span ID for nested spans
        shared_resource: Optional shared resource mock
        kind: The span kind (INTERNAL, SERVER, CLIENT, etc.)
        status_code: The span status code
        attributes: Optional span attributes

    Returns:
        A configured mock ReadableSpan
    """
    span = Mock(spec=ReadableSpan)
    span.name = name

    context = Mock(spec=SpanContext)
    context.trace_id = trace_id
    context.span_id = span_id
    context.trace_flags = TraceFlags(0x01)
    context.trace_state = TraceState()
    context.is_remote = False
    context.is_valid = True

    span.context = context
    span.get_span_context = Mock(return_value=context)

    if parent_span_id:
        span.parent = Mock(
            trace_id=trace_id,
            span_id=parent_span_id,
            trace_flags=TraceFlags(0x01),
            trace_state=TraceState(),
            is_remote=False,
            is_valid=True,
        )
    else:
        span.parent = None

    span.start_time = 1000000000000000000
    span.end_time = 1000000001000000000
    span.attributes = attributes or {}

    status_mock = Mock(spec=Status)
    status_mock.status_code = status_code
    status_mock.description = None
    span.status = status_mock

    span.events = []
    span.links = []
    span.kind = kind

    if shared_resource:
        span.resource = shared_resource
    else:
        span.resource = Mock()
        span.resource.attributes = {"service.name": "test-service"}

    span.instrumentation_scope = Mock()
    span.instrumentation_scope.name = "test-instrumentation"
    span.instrumentation_scope.version = "1.0.0"

    return span


def create_mock_span_with_time(
    name: str = "test_span",
    trace_id: int = 0x0102030405060708090A0B0C0D0E0F10,
    span_id: int = 0x0102030405060708,
) -> Mock:
    """Create a mock span with current time for retry/transport tests.

    This variant uses current time for start/end times and different
    default IDs for compatibility with existing retry tests.
    """
    span = Mock(spec=ReadableSpan)
    span.name = name

    context = Mock(spec=SpanContext)
    context.trace_id = trace_id
    context.span_id = span_id
    context.is_remote = False
    context.trace_flags = TraceFlags(0x01)
    context.trace_state = None

    span.context = context
    span.get_span_context = Mock(return_value=context)

    span.parent = None
    span.kind = SpanKind.SERVER
    span.start_time = int(time.time() * 1e9)
    span.end_time = int(time.time() * 1e9) + 1000000

    status_mock = Mock(spec=Status)
    status_mock.status_code = StatusCode.OK
    status_mock.description = None
    span.status = status_mock

    span.attributes = {}
    span.events = []
    span.links = []

    span.resource = Mock()
    span.resource.attributes = {"service.name": "test-service"}

    span.instrumentation_scope = Mock()
    span.instrumentation_scope.name = "test-scope"
    span.instrumentation_scope.version = "1.0.0"

    return span



"""Tests for Image content class."""

import tempfile
from pathlib import Path

import httpx
import pytest
from pytest_httpserver import HTTPServer

from mirascope.llm.content.image import (
    MAX_IMAGE_SIZE,
    Base64ImageSource,
    Image,
    URLImageSource,
)


@pytest.fixture
def image_data() -> dict[str, bytes]:
    """Provide sample image data for different formats (minimum 12 bytes each)."""
    return {
        "png": b"\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR",
        "jpeg": b"\xff\xd8\xff\xe0\x00\x10JFIF\x00\x01\x01",
        "webp": b"RIFF\x00\x00\x00\x00WEBP",
        "gif87a": b"GIF87a\x00\x00\x00\x00\x00\x00",
        "gif89a": b"GIF89a\x00\x00\x00\x00\x00\x00",
        "heic_ftypheic": b"\x00\x00\x00\x18ftypheic",
        "heic_ftypheix": b"\x00\x00\x00\x18ftypheix",
        "heif_ftypmif1": b"\x00\x00\x00\x18ftypmif1",
        "heif_ftypmsf1": b"\x00\x00\x00\x18ftypmsf1",
        "heif_ftyphevc": b"\x00\x00\x00\x18ftyphevc",
        "heif_ftyphevx": b"\x00\x00\x00\x18ftyphevx",
        "unsupported": b"random unsupported data",
    }


IMAGE_FORMAT_TESTS = [
    ("png", "image/png"),
    ("jpeg", "image/jpeg"),
    ("webp", "image/webp"),
    ("gif87a", "image/gif"),
    ("gif89a", "image/gif"),
    ("heic_ftypheic", "image/heic"),
    ("heic_ftypheix", "image/heic"),
    ("heif_ftypmif1", "image/heif"),
    ("heif_ftypmsf1", "image/heif"),
    ("heif_ftyphevc", "image/heif"),
    ("heif_ftyphevx", "image/heif"),
]


class TestImageFromUrl:
    """Tests for Image.from_url class method."""

    def test_from_url_creates_url_reference(self) -> None:
        """Test creating an Image reference from URL without downloading."""
        url = "https://example.com/image.png"
        image = Image.from_url(url)

        assert isinstance(image.source, URLImageSource)
        assert image.source.url == url
        assert image.source.type == "url_image_source"


class TestImageDownload:
    """Tests for Image.download class method."""

    @pytest.mark.parametrize("format_name,expected_mime", IMAGE_FORMAT_TESTS)
    def test_download_detects_mime_type_from_magic_bytes(
        self,
        httpserver: HTTPServer,
        image_data: dict[str, bytes],
        format_name: str,
        expected_mime: str,
    ) -> None:
        """Test that download() detects MIME type from magic bytes."""
        httpserver.expect_request("/image").respond_with_data(image_data[format_name])
        url = httpserver.url_for("/image")

        image = Image.download(url)

        assert isinstance(image.source, Base64ImageSource)
        assert image.source.type == "base64_image_source"
        assert image.source.mime_type == expected_mime
        assert len(image.source.data) > 0

    def test_download_follows_redirects(
        self, httpserver: HTTPServer, image_data: dict[str, bytes]
    ) -> None:
        """Test that download follows redirects."""
        httpserver.expect_request("/redirect").respond_with_data(
            "", status=302, headers={"Location": "/image"}
        )
        httpserver.expect_request("/image").respond_with_data(image_data["png"])
        url = httpserver.url_for("/redirect")

        image = Image.download(url)

        assert isinstance(image.source, Base64ImageSource)
        assert image.source.mime_type == "image/png"

    def test_download_raises_on_unsupported_format(
        self, httpserver: HTTPServer, image_data: dict[str, bytes]
    ) -> None:
        """Test that download raises ValueError for unsupported image format."""
        httpserver.expect_request("/image").respond_with_data(image_data["unsupported"])
        url = httpserver.url_for("/image")

        with pytest.raises(ValueError, match="Unsupported image type"):
            Image.download(url)

    def test_download_raises_on_http_error(self, httpserver: HTTPServer) -> None:
        """Test that download raises on HTTP errors."""
        httpserver.expect_request("/image").respond_with_data("", status=404)
        url = httpserver.url_for("/image")

        with pytest.raises(httpx.HTTPStatusError):
            Image.download(url)

    def test_download_enforces_size_limit(self, httpserver: HTTPServer) -> None:
        """Test that download enforces the size limit."""
        large_data = b"\xff\xd8\xff" + b"x" * (MAX_IMAGE_SIZE + 1)
        httpserver.expect_request("/image").respond_with_data(large_data)
        url = httpserver.url_for("/image")

        with pytest.raises(ValueError, match="exceeds maximum allowed size"):
            Image.download(url)

    def test_download_respects_custom_size_limit(
        self, httpserver: HTTPServer, image_data: dict[str, bytes]
    ) -> None:
        """Test that download respects custom max_size parameter."""
        httpserver.expect_request("/image").respond_with_data(image_data["png"])
        url = httpserver.url_for("/image")

        with pytest.raises(ValueError, match="exceeds maximum allowed size"):
            Image.download(url, max_size=10)


class TestImageDownloadAsync:
    """Tests for Image.download_async class method."""

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "format_name,expected_mime",
        [
            ("png", "image/png"),
            ("jpeg", "image/jpeg"),
            ("webp", "image/webp"),
            ("gif89a", "image/gif"),
        ],
    )
    async def test_download_async_detects_mime_type(
        self,
        httpserver: HTTPServer,
        image_data: dict[str, bytes],
        format_name: str,
        expected_mime: str,
    ) -> None:
        """Test that download_async() detects MIME type from magic bytes."""
        httpserver.expect_request("/image").respond_with_data(image_data[format_name])
        url = httpserver.url_for("/image")

        image = await Image.download_async(url)

        assert isinstance(image.source, Base64ImageSource)
        assert image.source.mime_type == expected_mime

    @pytest.mark.asyncio
    async def test_download_async_enforces_size_limit(
        self, httpserver: HTTPServer
    ) -> None:
        """Test that download_async enforces the size limit."""
        large_data = b"\xff\xd8\xff" + b"x" * (MAX_IMAGE_SIZE + 1)
        httpserver.expect_request("/image").respond_with_data(large_data)
        url = httpserver.url_for("/image")

        with pytest.raises(ValueError, match="exceeds maximum allowed size"):
            await Image.download_async(url)

    @pytest.mark.asyncio
    async def test_download_async_raises_on_unsupported_format(
        self, httpserver: HTTPServer, image_data: dict[str, bytes]
    ) -> None:
        """Test that download_async raises ValueError for unsupported format."""
        httpserver.expect_request("/image").respond_with_data(image_data["unsupported"])
        url = httpserver.url_for("/image")

        with pytest.raises(ValueError, match="Unsupported image type"):
            await Image.download_async(url)


class TestImageFromFile:
    """Tests for Image.from_file class method."""

    @pytest.mark.parametrize("data_key,expected_mime", IMAGE_FORMAT_TESTS)
    def test_from_file_detects_mime_type(
        self, image_data: dict[str, bytes], data_key: str, expected_mime: str
    ) -> None:
        """Test that from_file detects MIME type from magic bytes (not extension)."""
        with tempfile.NamedTemporaryFile(suffix=".unknown", delete=False) as f:
            f.write(image_data[data_key])
            temp_path = f.name

        try:
            image = Image.from_file(temp_path)

            assert isinstance(image.source, Base64ImageSource)
            assert image.source.mime_type == expected_mime
        finally:
            Path(temp_path).unlink()

    def test_from_file_raises_on_unsupported_format(
        self, image_data: dict[str, bytes]
    ) -> None:
        """Test that from_file raises ValueError for unsupported format."""
        with tempfile.NamedTemporaryFile(suffix=".bin", delete=False) as f:
            f.write(image_data["unsupported"])
            temp_path = f.name

        try:
            with pytest.raises(ValueError, match="Unsupported image type"):
                Image.from_file(temp_path)
        finally:
            Path(temp_path).unlink()

    def test_from_file_not_found(self) -> None:
        """Test FileNotFoundError is raised for non-existent file."""
        with pytest.raises(FileNotFoundError):
            Image.from_file("/nonexistent/path/to/file.png")

    def test_from_file_enforces_size_limit(self, image_data: dict[str, bytes]) -> None:
        """Test that from_file enforces the size limit."""
        large_data = b"\xff\xd8\xff" + b"x" * (MAX_IMAGE_SIZE + 1)
        with tempfile.NamedTemporaryFile(suffix=".jpg", delete=False) as f:
            f.write(large_data)
            temp_path = f.name

        try:
            with pytest.raises(ValueError, match="exceeds maximum allowed size"):
                Image.from_file(temp_path)
        finally:
            Path(temp_path).unlink()

    def test_from_file_respects_custom_size_limit(
        self, image_data: dict[str, bytes]
    ) -> None:
        """Test that from_file respects custom max_size parameter."""
        with tempfile.NamedTemporaryFile(suffix=".png", delete=False) as f:
            f.write(image_data["png"])
            temp_path = f.name

        try:
            # Should fail with a very small limit
            with pytest.raises(ValueError, match="exceeds maximum allowed size"):
                Image.from_file(temp_path, max_size=10)
        finally:
            Path(temp_path).unlink()


class TestImageFromBytes:
    """Tests for Image.from_bytes class method."""

    @pytest.mark.parametrize("data_key,expected_mime", IMAGE_FORMAT_TESTS)
    def test_from_bytes_detects_mime_type(
        self, image_data: dict[str, bytes], data_key: str, expected_mime: str
    ) -> None:
        """Test that from_bytes detects MIME type from magic bytes."""
        image = Image.from_bytes(image_data[data_key])

        assert isinstance(image.source, Base64ImageSource)
        assert image.source.mime_type == expected_mime

    def test_from_bytes_raises_on_unsupported_format(
        self, image_data: dict[str, bytes]
    ) -> None:
        """Test that from_bytes raises ValueError for unsupported format."""
        with pytest.raises(ValueError, match="Unsupported image type"):
            Image.from_bytes(image_data["unsupported"])

    def test_from_bytes_raises_on_data_too_small(self) -> None:
        """Test that from_bytes raises ValueError for data that's too small."""
        with pytest.raises(ValueError, match="Image data too small to determine type"):
            Image.from_bytes(b"short")

    def test_from_bytes_enforces_size_limit(self) -> None:
        """Test that from_bytes enforces the size limit."""
        large_data = b"\xff\xd8\xff" + b"x" * (MAX_IMAGE_SIZE + 1)

        with pytest.raises(ValueError, match="exceeds maximum allowed size"):
            Image.from_bytes(large_data)

    def test_from_bytes_respects_custom_size_limit(
        self, image_data: dict[str, bytes]
    ) -> None:
        """Test that from_bytes respects custom max_size parameter."""
        with pytest.raises(ValueError, match="exceeds maximum allowed size"):
            Image.from_bytes(image_data["png"], max_size=10)



"""End-to-end tests for a LLM call with all params set."""

from typing import get_type_hints

import pytest

from mirascope import llm
from tests.e2e.conftest import E2E_MODEL_IDS
from tests.utils import (
    Snapshot,
    response_snapshot_dict,
    snapshot_test,
)

PARAMS_MODEL_IDS: list[llm.ModelId] = [
    *E2E_MODEL_IDS,
    # Include reasoning models to verify unsupported params (temperature, top_p) are logged
    "openai/gpt-5-mini:completions",
    "openai/gpt-5:completions",
]

# ============= ALL PARAMS TESTS =============
ALL_PARAMS: llm.Params = {
    "temperature": 0.7,
    "max_tokens": 500,
    "top_p": 0.3,
    "top_k": 50,
    "seed": 42,
    "stop_sequences": ["4242"],
    "thinking": {
        "level": "none",
        "encode_thoughts_as_text": False,
        "include_thoughts": False,
    },
}


def test_all_params_includes_every_param() -> None:
    """Verify that ALL_PARAMS includes every parameter defined in Params."""
    params_keys = set(get_type_hints(llm.Params).keys())
    all_params_keys = set(ALL_PARAMS.keys())
    assert params_keys == all_params_keys, (
        f"ALL_PARAMS is missing parameters: {params_keys - all_params_keys}"
    )


@pytest.mark.parametrize("model_id", PARAMS_MODEL_IDS)
@pytest.mark.vcr
def test_call_with_params(
    model_id: llm.ModelId,
    snapshot: Snapshot,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test synchronous call with all parameters to verify param handling and logging."""

    @llm.call(model_id, **ALL_PARAMS)
    def add_numbers(a: int, b: int) -> str:
        return f"What is {a} + {b}?"

    with snapshot_test(snapshot, caplog) as snap:
        response = add_numbers(4200, 42)
        snap["response"] = (response_snapshot_dict(response),)



"""End-to-end tests for structured output with validation retry using response.validate()."""

import pytest
from pydantic import BaseModel, field_validator

from mirascope import llm
from tests.e2e.conftest import STRUCTURED_OUTPUT_MODEL_IDS
from tests.utils import (
    Snapshot,
    snapshot_test,
)


class LuckyNumber(BaseModel):
    """A lucky number chosen by the model."""

    value: int

    @field_validator("value")
    @classmethod
    def must_be_173(cls, v: int) -> int:
        """Validate that the number is exactly 173.

        This constraint is NOT in the schema, so the LLM won't know about it
        on the first attempt, forcing a validation retry.
        """
        if v != 173:
            raise ValueError(f"The number must be exactly 173, but got {v}")
        return v


# ============= SYNC TESTS =============


@pytest.mark.parametrize("model_id", STRUCTURED_OUTPUT_MODEL_IDS)
@pytest.mark.vcr
def test_validation_retry_sync(
    model_id: llm.ModelId,
    snapshot: Snapshot,
) -> None:
    """Test that validate() retries on parse error and succeeds after LLM correction."""

    @llm.call(
        model_id,
        format=llm.format(LuckyNumber, mode="json"),
    )
    def pick_number() -> str:
        return "Pick a random number between 1 and 1000"

    with snapshot_test(snapshot) as snap:
        response = pick_number()

        # validate() should retry when the number isn't 173
        number, final_response = response.validate(max_retries=1)

        snap.set_response(final_response)

        assert number is not None
        assert number.value == 173


@pytest.mark.parametrize("model_id", STRUCTURED_OUTPUT_MODEL_IDS)
@pytest.mark.vcr
def test_validation_retry_sync_context(
    model_id: llm.ModelId,
    snapshot: Snapshot,
) -> None:
    """Test that validate() works with context responses."""

    @llm.call(
        model_id,
        format=llm.format(LuckyNumber, mode="json"),
    )
    def pick_number(ctx: llm.Context[None]) -> str:
        return "Pick a random number between 1 and 1000"

    ctx = llm.Context(deps=None)
    with snapshot_test(snapshot) as snap:
        response = pick_number(ctx)

        number, final_response = response.validate(ctx, max_retries=1)

        snap.set_response(final_response)

        assert number is not None
        assert number.value == 173


# ============= ASYNC TESTS =============


@pytest.mark.parametrize("model_id", STRUCTURED_OUTPUT_MODEL_IDS)
@pytest.mark.vcr
@pytest.mark.asyncio
async def test_validation_retry_async(
    model_id: llm.ModelId,
    snapshot: Snapshot,
) -> None:
    """Test that async validate() retries on parse error and succeeds after LLM correction."""

    @llm.call(
        model_id,
        format=llm.format(LuckyNumber, mode="json"),
    )
    async def pick_number() -> str:
        return "Pick a random number between 1 and 1000"

    with snapshot_test(snapshot) as snap:
        response = await pick_number()

        number, final_response = await response.validate(max_retries=1)

        snap.set_response(final_response)

        assert number is not None
        assert number.value == 173


@pytest.mark.parametrize("model_id", STRUCTURED_OUTPUT_MODEL_IDS)
@pytest.mark.vcr
@pytest.mark.asyncio
async def test_validation_retry_async_context(
    model_id: llm.ModelId,
    snapshot: Snapshot,
) -> None:
    """Test that async validate() works with context responses."""

    @llm.call(
        model_id,
        format=llm.format(LuckyNumber, mode="json"),
    )
    async def pick_number(ctx: llm.Context[None]) -> str:
        return "Pick a random number between 1 and 1000"

    ctx = llm.Context(deps=None)
    with snapshot_test(snapshot) as snap:
        response = await pick_number(ctx)

        number, final_response = await response.validate(ctx, max_retries=1)

        snap.set_response(final_response)

        assert number is not None
        assert number.value == 173


# ============= STREAM TESTS =============


@pytest.mark.parametrize("model_id", STRUCTURED_OUTPUT_MODEL_IDS)
@pytest.mark.vcr
def test_validation_retry_stream(
    model_id: llm.ModelId,
    snapshot: Snapshot,
) -> None:
    """Test that stream validate() retries on parse error."""

    @llm.call(
        model_id,
        format=llm.format(LuckyNumber, mode="json"),
    )
    def pick_number() -> str:
        return "Pick a random number between 1 and 1000"

    with snapshot_test(snapshot) as snap:
        response = pick_number.stream()

        # validate() consumes the stream and retries if needed
        number, final_response = response.validate(max_retries=1)

        snap.set_response(final_response)

        assert number is not None
        assert number.value == 173


@pytest.mark.parametrize("model_id", STRUCTURED_OUTPUT_MODEL_IDS)
@pytest.mark.vcr
def test_validation_retry_stream_context(
    model_id: llm.ModelId,
    snapshot: Snapshot,
) -> None:
    """Test that stream validate() works with context responses."""

    @llm.call(
        model_id,
        format=llm.format(LuckyNumber, mode="json"),
    )
    def pick_number(ctx: llm.Context[None]) -> str:
        return "Pick a random number between 1 and 1000"

    ctx = llm.Context(deps=None)
    with snapshot_test(snapshot) as snap:
        response = pick_number.stream(ctx)

        number, final_response = response.validate(ctx, max_retries=1)

        snap.set_response(final_response)

        assert number is not None
        assert number.value == 173


# ============= ASYNC STREAM TESTS =============


@pytest.mark.parametrize("model_id", STRUCTURED_OUTPUT_MODEL_IDS)
@pytest.mark.vcr
@pytest.mark.asyncio
async def test_validation_retry_async_stream(
    model_id: llm.ModelId,
    snapshot: Snapshot,
) -> None:
    """Test that async stream validate() retries on parse error."""

    @llm.call(
        model_id,
        format=llm.format(LuckyNumber, mode="json"),
    )
    async def pick_number() -> str:
        return "Pick a random number between 1 and 1000"

    with snapshot_test(snapshot) as snap:
        response = await pick_number.stream()

        number, final_response = await response.validate(max_retries=1)

        snap.set_response(final_response)

        assert number is not None
        assert number.value == 173


@pytest.mark.parametrize("model_id", STRUCTURED_OUTPUT_MODEL_IDS)
@pytest.mark.vcr
@pytest.mark.asyncio
async def test_validation_retry_async_stream_context(
    model_id: llm.ModelId,
    snapshot: Snapshot,
) -> None:
    """Test that async stream validate() works with context responses."""

    @llm.call(
        model_id,
        format=llm.format(LuckyNumber, mode="json"),
    )
    async def pick_number(ctx: llm.Context[None]) -> str:
        return "Pick a random number between 1 and 1000"

    ctx = llm.Context(deps=None)
    with snapshot_test(snapshot) as snap:
        response = await pick_number.stream(ctx)

        number, final_response = await response.validate(ctx, max_retries=1)

        snap.set_response(final_response)

        assert number is not None
        assert number.value == 173



from mirascope import llm


def recommend_book(model_id: llm.ModelId, genre: str):
    model = llm.Model(model_id)
    return model.call(f"Please recommend a book in {genre}.")


response = recommend_book("anthropic/claude-haiku-4-5", "fantasy")
print(response.pretty())



"""Explicit span management utilities for `mirascope.ops`."""

from __future__ import annotations

import logging
from contextvars import Token
from typing import TYPE_CHECKING, Any

from opentelemetry import context as otel_context, trace as otel_trace
from opentelemetry.trace import (
    SpanContext,
    Status,
    StatusCode,
    format_span_id,
    format_trace_id,
)
from opentelemetry.util.types import AttributeValue

from .session import current_session
from .utils import json_dumps

if TYPE_CHECKING:
    from opentelemetry.context import Context

logger = logging.getLogger("mirascope.ops")
_warned_noop = False


class Span:
    """Context-managed span for explicit tracing.

    Creates a child span within the current trace context. Acts as a no-op
    if tracing is not configured.
    """

    def __init__(self, name: str, **attributes: AttributeValue) -> None:
        """Initialize a new span with the given name.

        Args:
            name: Name for the span.
            **attributes: Initial attributes to set on the span.
        """
        self._name = name
        self._initial_attributes = attributes
        self._span: otel_trace.Span | None = None
        self._token: Token[Context] | None = None
        self._is_noop = True
        self._finished = False

    def __enter__(self) -> Span:
        """Enter the span context.

        Returns:
            This span instance for use within the context.
        """
        tracer = otel_trace.get_tracer("mirascope.ops")
        self._span = tracer.start_span(self._name)

        if self._span.__class__.__name__ == "NonRecordingSpan":
            self._is_noop = True
            self._span = None
            global _warned_noop
            if not _warned_noop:
                logger.warning(
                    f"mirascope tracing is not configured; Span('{self._name}') is a no-op."
                )
                _warned_noop = True
        else:
            self._is_noop = False
            self._span.set_attribute("mirascope.type", "trace")

            session_ctx = current_session()
            if session_ctx is not None:
                self._span.set_attribute("mirascope.ops.session.id", session_ctx.id)
                if session_ctx.attributes is not None:
                    self._span.set_attribute(
                        "mirascope.ops.session.attributes",
                        json_dumps(session_ctx.attributes),
                    )

            self._token = otel_context.attach(
                otel_trace.set_span_in_context(self._span)
            )

            if self._initial_attributes:
                self.set(**self._initial_attributes)

        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        tb: Any,  # noqa: ANN401
    ) -> None:
        """Exit the span context.

        Args:
            exc_type: Exception type if an exception was raised.
            exc: Exception instance if an exception was raised.
            tb: Traceback if an exception was raised.
        """
        if self._span and not self._finished:
            if exc is not None:
                self._span.record_exception(exc)
                self._span.set_status(Status(StatusCode.ERROR))
            self.finish()

    def set(self, **attributes: AttributeValue) -> None:
        """Set attributes on the current span.

        Args:
            **attributes: Key-value pairs to set as span attributes.
        """
        if self._span and not self._finished:
            for key, value in attributes.items():
                self._span.set_attribute(key, value)

    def event(self, name: str, **attributes: AttributeValue) -> None:
        """Record an event within the span.

        Args:
            name: Name of the event.
            **attributes: Event attributes as key-value pairs.
        """
        if self._span and not self._finished:
            self._span.add_event(name, attributes=attributes)

    def debug(self, message: str, **additional_attributes: AttributeValue) -> None:
        """Log a debug message within the span.

        Args:
            message: Debug message text.
            **additional_attributes: Additional structured attributes for the log entry.
        """
        self.event("debug", level="debug", message=message, **additional_attributes)

    def info(self, message: str, **additional_attributes: AttributeValue) -> None:
        """Log an info message within the span.

        Args:
            message: Info message text.
            **additional_attributes: Additional structured attributes for the log entry.
        """
        self.event("info", level="info", message=message, **additional_attributes)

    def warning(self, message: str, **additional_attributes: AttributeValue) -> None:
        """Log a warning message within the span.

        Args:
            message: Warning message text.
            **additional_attributes: Additional structured attributes for the log entry.
        """
        self.event("warning", level="warning", message=message, **additional_attributes)

    def error(self, message: str, **additional_attributes: AttributeValue) -> None:
        """Log an error message within the span.

        Args:
            message: Error message text.
            **additional_attributes: Additional structured attributes for the log entry.
        """
        self.event("error", level="error", message=message, **additional_attributes)
        if self._span and not self._finished:
            self._span.set_status(Status(StatusCode.ERROR))

    def critical(self, message: str, **additional_attributes: AttributeValue) -> None:
        """Log a critical message within the span.

        Args:
            message: Critical message text.
            **additional_attributes: Additional structured attributes for the log entry.
        """
        self.event("critical", level="error", message=message, **additional_attributes)
        if self._span and not self._finished:
            self._span.set_status(Status(StatusCode.ERROR))

    def finish(self) -> None:
        """Explicitly finish the span."""
        if not self._finished:
            self._finished = True
            if self._span:
                self._span.end()
            if self._token:
                otel_context.detach(self._token)
                self._token = None

    @property
    def span_id(self) -> str | None:
        """Get the span ID if available.

        Returns:
            The span ID or None if not available.
        """
        if span_context := self.span_context:
            return format_span_id(span_context.span_id)
        return None

    @property
    def trace_id(self) -> str | None:
        if span_context := self.span_context:
            return format_trace_id(span_context.trace_id)
        return None

    @property
    def is_noop(self) -> bool:
        """Check if this is a no-op span.

        Returns:
            True if tracing is disabled, False otherwise.
        """
        return self._is_noop

    @property
    def span_context(self) -> SpanContext | None:
        """Return the span context if available."""
        if otel_span := self._span:
            return otel_span.get_span_context()
        return None


def span(name: str, **attributes: AttributeValue) -> Span:
    """Create a new span context manager.

    Args:
        name: Name for the new span.
        **attributes: Initial attributes to set on the span.

    Returns:
        A Span context manager that creates a child span when entered.
    """
    return Span(name, **attributes)



from mirascope import llm


@llm.call("mlx-community/Qwen3-8B-4bit-DWQ-053125")
def recommend_book(genre: str):
    return f"Recommend a {genre} book."


def main():
    response: llm.Response = recommend_book("fantasy")
    print(response.pretty())


main()



from inline_snapshot import snapshot

from mirascope.llm import (
    AssistantMessage,
    Text,
    UserMessage,
)

test_snapshot = snapshot(
    {
        "response": {
            "provider_id": "google",
            "model_id": "google/gemini-3-pro-preview",
            "provider_model_name": "gemini-3-pro-preview",
            "finish_reason": None,
            "messages": [
                UserMessage(
                    content=[Text(text="Answer this question: What is 2 + 2?")]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text="""\
{
"integer_a": 2,
"integer_b": 2,
"answer": 4
}\
"""
                        )
                    ],
                    provider_id="google",
                    model_id="google/gemini-3-pro-preview",
                    provider_model_name="gemini-3-pro-preview",
                    raw_message={
                        "parts": [
                            {
                                "media_resolution": None,
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_call": None,
                                "function_response": None,
                                "inline_data": None,
                                "text": """\
{
"integer_a": 2,
"integer_b": 2,
"answer": 4
}\
""",
                                "thought": None,
                                "thought_signature": None,
                                "video_metadata": None,
                            },
                            {
                                "media_resolution": None,
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_call": None,
                                "function_response": None,
                                "inline_data": None,
                                "text": "",
                                "thought": None,
                                "thought_signature": b"\x12\xe6\x08\n\xe3\x08\x01r\xc8\xda|$w\xef'x\x04\xe5{\xb6g*\x8f6\x86D1\x13\xad\x8b\xcd#\x1e\xd5\xc7\x96\xaceat\xcbTE\xc6U\xf0\x97\x8e\x87\x15,[\x8b\xb3\xf0\x8b\xd188,\xc1\xfaFg\xe4\xe4=\xf6\xb1\x90\x81lT\xe3\x88d\xa2\x19\x96x\xf2\x9d\xe1\xd5\x10.u\xf8\x19\x88?\xee\xd9Q\xb3\x05\xf6\x08e\x88<\xa6\xf7\x13\x03\xd69r>\x88{8^\xb4u#\x95\xa8U1\x9d\xbf0\x00\xddW\x8b@G\xcb\xab5\x88\x1f*\xd12@\xdc\xa3\xea\x02W\xba\xf5\xda\xc0\xc5\xe4\x8eZX\x98\n\x958Y'\xcb\xe6\x9c\xd8\xd6\x0bbV\xa6\r\x94\x8e\xdf\xd5\"9\xe0\xce\x055\x06\xa0K\x14\x1b\x9aj\xc0\x01mUlf\r\xe7\xe9\x0b\x8b\xdd\xc9\x94\x81K\xc7\x85\xb7z\x10\xdc|\xd3\xe2n\xd3\xe7\xe8e\xcc\xfe\xb5p\xb8It\xe7\x8d\xc8\xef\x14\x11\xa9I\xaeM\x15\xebD\xfd\x12\x93\xf7mH\x01B\xb88\x1d\xcc\xdb\x0f\n+\xc7(\x19ap_\xb1\xd2\xd4b\x1blh'\xa1\xe9\xf3'\x105S)j\xd0\xc07\xe9\t\xb4\xe7\x92:\xbbf$x\xcf\xc6\xa9p\xf8\x92\x7f;\xab\xd96\x84\xc4\x7f\xad\x10zm\xc4\x88T+\xb3\xe3\xe09d\xcb&\x91\xc0\x0b\xb4\x8e\xd3W\xe2\xdb\xe7.\xbf\xaaIy\x90d\xd5\x13f\x81\x05\\\x19\xde>\xa8\xf2$.|\x8a\xa9\xa2Pp\x0c6g\n\xa7\x82r\x8c\x12\xcb0i?\xe9\x129t\x85\xad\xa7\x16W\xf7j\xf5\xc8+\x0c\xd5J\x10\x95\xe0P\xb9\x89\xeb\xe0\xd3L\xf35a\xae\x9c9`\xa6\\iXoq\xa9\xd6\xf2\xc0\xfe\x81\x99\x07\x93\xce\xe6\xb0t\xe2\x92n1\x03\x0c\x8d\x05x\xf1\xcc\xba\x080\x90O@k\xe18\x84\xa6q\x8a}\x81_\n\x95\x89\xd4\xe6\x04\xe5\xfa\xa4\xc5\x13\xad\x0c\xfe\xaa\x9b\x90\\\nd\xa8\xdfi\xff-\x9e\xa7\x104\x9ar3?\xf8B}B\x89FY\x0f\xb4H\xfd\"@G\xcb\xcc\xbe!(\r\x95\xc9F\xc3\x97\x04X\x90+\xe2#/c\xd7\xd1\x13\xdbX\x03(\xc0\xf5u\xc4\xffY\xd4\x99\x02\x08\x87\xcb\x89 g\xc5\xa1=+\x15F\x9ar\xd6\x0ev\x1e\xe0\xae\x9e\xd6\xea\x1e)\x1e\x06\xe0\x82#\x0b\x17\xfc\xf7C\x8f]\x02\x89\xa9\r\x00\xe0|M\x9d\xf3n\xc6\xfa\x8d?\x03\xd0\xac\xcd\xe9C:\x8f0P\xfa)\xb8\xbd!sY\xfd =r7\xab\xd4`\x86J\xa9+\xf8A\xba\xda\xda\x88r\xeb\xa62\xe7\xaf\xba$\x17lPG^\xea&E\xa3\xef\xf9\xcc\xbeE\xa2a\xbe\xac\x82\xc8\xff\x07\xbe6\xf9\xae6\x8c\xaf\xc4(\x96\x7f&\xfaZ_\xa6~\x1b\x99\xde\xc2\x1e\xcc\xa1\x05\x1d\xee\x17\xc3\xb9\xdf\xa0\xd6\x05\x11\xbc\x9c\xe0\xd4\xee8X\xe8\xa6\xfd\xc3\x14-\xb7\xf1@rfp\xf9\x90\x10\xb0\xc4\xaf\xf6\x98G\xcd\x9c#\x90\x96c<\xe2<\x06\x1b\x1d\x985t\x98(\xb1\x0e\xae\xf9)\xb0C\xb5\xd6\xa3\xe0+\x0bB\xdf\x9a\x02't\xc2\xc0\x04+\xd3\xc1,\xa6\xdf\xee\xaf\xb4f\xbb{\x97|f\xde\xf3\xd3\xb6r\xfenJ\x17T2\xf6\xcf\x18\xc7v]\xe0\xb5\xa9x\xfd\xb4<^y\xe1\xc4\x06I\x81\xdd9v\x82\xe7/\x10\xa5\x86[\xb6\xc2N\xb5\x1d\x82f\xb9R\xdf\xd2\xfd\xe5\xfcvc\x1fN\xd3\xc9X\x90\xa2\x11\x92\x01*\xbb\x1d[\xb3\xb6\xa1\xa2*\xc2\x8fV\xc3\xfc\xb5]\x00>\xe6\x1c\xee;^\r9\xf6\xe4\x96^\xb9\xdf\x19\xef\x17(@\xa8\xa3\xc5\x19\xadU\xadJ\xea\xfae\xb4\x196\xe0\x88\xdb\xd2\xcc\x86\xc1<\xf4q\r2\xaaI]\xd4\"R\x83k\xf9\x92a\xb8\xc3\xbf\x1a\x19xI\x12\xa2\xb0{\xdcW{\x13\xae\xe1\xae\x16T?4l\xe7\x0b\xd2f\xc8\n\xdd?t)~0\xf7T{y\nB\xf9\xcc\x81rw,C\x8c\xcc\xf31\x89F\xbd{\x01\xef\xac6\xea\x9c#FP\xd8\xee\xe5\xb6\xb3\xf9\xe2\x04y\xb9\xe7\x92'\x11?\x9aG\x1c\x03z\x90!\x11\x11\x17CG\xc8v\xc6v\x1b\x03d}\xf6\xfb\xc3\xdc\x8dP\x15\xc2\xcb\x95\x97\x98V\x01\xa3X\x11\x1cC}o\x90\xc3\xfe\xe3\xb6DTxx\xa1#>\xcc\xbd\xdc\xcb5C!\x89%\xc6\x1f\xa0\x1f\x0f\xd6\xb7\x18\xda\x93a\t\xd1\xa8Hy\x92'\xd0\x18\xcc\xe8\xfbz\x98x\xc1i\x89\x86mub\xbf\x9d\xdfkH\x06\xa7\xdf\xc8\xe1\x08f\r\xc9/zb{\xb7\xa27$\x9e\x86\xcd\"\xd4r\xda\xf5\xcd\xa96X?\xd1\x1f\xa2\xf6\xeb\x92\x97\x16\xebf\xec@U\xbb>\xc9\xac\x0f",
                                "video_metadata": None,
                            },
                        ],
                        "role": "model",
                    },
                ),
            ],
            "format": {
                "name": "IntegerAdditionResponse",
                "description": "A simple response for testing.",
                "schema": {
                    "description": "A simple response for testing.",
                    "properties": {
                        "integer_a": {"title": "Integer A", "type": "integer"},
                        "integer_b": {"title": "Integer B", "type": "integer"},
                        "answer": {"title": "Answer", "type": "integer"},
                    },
                    "required": ["integer_a", "integer_b", "answer"],
                    "title": "IntegerAdditionResponse",
                    "type": "object",
                },
                "mode": "strict",
                "formatting_instructions": None,
            },
            "tools": [],
            "usage": {
                "input_tokens": 13,
                "output_tokens": 27,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 284,
                "provider_tool_usage": None,
                "raw": "None",
                "total_tokens": 40,
            },
            "n_chunks": 3,
        }
    }
)



"""End-to-end tests for structured output."""

import re
import xml.etree.ElementTree as ET

import pytest
from pydantic import BaseModel

from mirascope import llm
from tests.utils import (
    Snapshot,
    snapshot_test,
)

# This behavior is largely model-agnostic, so we will test on just one model
# to reduce test bloat.
MODEL_IDS = ["anthropic/claude-haiku-4-5"]


class Book(BaseModel):
    title: str
    rating: int


@pytest.mark.parametrize("model_id", MODEL_IDS)
@pytest.mark.vcr
def test_structured_output_with_output_parser(
    model_id: llm.ModelId,
    snapshot: Snapshot,
) -> None:
    """Test using output parser"""

    instructions = (
        "Return book as XML: <book><title>Mistborn</title><rating>7</rating></book>"
    )

    @llm.output_parser(formatting_instructions=instructions)
    def parse_book(response: llm.AnyResponse) -> Book:
        text = "".join(t.text for t in response.texts)
        # Strip markdown code fences if present
        xml_match = re.search(r"<book>.*</book>", text, re.DOTALL)
        xml_text = xml_match.group(0) if xml_match else text
        root = ET.fromstring(xml_text)
        title = root.findtext("title") or ""
        rating = int(root.findtext("rating") or "0")
        return Book(title=title, rating=rating)

    @llm.call(model_id, format=parse_book)
    def recommend_book() -> str:
        return "The Name of the Wind, Rating 7"

    with snapshot_test(snapshot) as snap:
        response = recommend_book()
        snap.set_response(response)

        book: Book = response.parse()
        assert book.title == "The Name of the Wind"
        assert book.rating == 7



"""Tests for RetryConfig validation."""

import pytest

from mirascope import llm

from .mock_provider import MockProvider


def test_max_retries_negative_raises_value_error() -> None:
    """Test that negative max_retries raises ValueError."""
    with pytest.raises(ValueError, match="max_retries must be non-negative"):
        llm.retry(max_retries=-1)


def test_max_retries_zero_is_valid(mock_provider: MockProvider) -> None:
    """Test that max_retries=0 is valid (no retries, just one attempt)."""

    @llm.retry(max_retries=0)
    @llm.call("mock/test-model")
    def greet() -> str:
        return "Hello"

    response = greet()
    assert response.model.retry_config.max_retries == 0


def test_initial_delay_negative_raises_value_error() -> None:
    """Test that negative initial_delay raises ValueError."""
    with pytest.raises(ValueError, match="initial_delay must be non-negative"):
        llm.retry(initial_delay=-1.0)


def test_max_delay_negative_raises_value_error() -> None:
    """Test that negative max_delay raises ValueError."""
    with pytest.raises(ValueError, match="max_delay must be non-negative"):
        llm.retry(max_delay=-1.0)


def test_backoff_multiplier_less_than_one_raises_value_error() -> None:
    """Test that backoff_multiplier < 1 raises ValueError."""
    with pytest.raises(ValueError, match="backoff_multiplier must be >= 1"):
        llm.retry(backoff_multiplier=0.5)


def test_jitter_out_of_range_raises_value_error() -> None:
    """Test that jitter outside [0, 1] raises ValueError."""
    with pytest.raises(ValueError, match="jitter must be between 0.0 and 1.0"):
        llm.retry(jitter=-0.1)
    with pytest.raises(ValueError, match="jitter must be between 0.0 and 1.0"):
        llm.retry(jitter=1.1)



# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
import typing_extensions
from ...core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from ...core.serialization import FieldMetadata
from .traces_create_request_resource_spans_item_scope_spans_item_spans_item_attributes_item_value_array_value import (
    TracesCreateRequestResourceSpansItemScopeSpansItemSpansItemAttributesItemValueArrayValue,
)
from .traces_create_request_resource_spans_item_scope_spans_item_spans_item_attributes_item_value_int_value import (
    TracesCreateRequestResourceSpansItemScopeSpansItemSpansItemAttributesItemValueIntValue,
)
from .traces_create_request_resource_spans_item_scope_spans_item_spans_item_attributes_item_value_kvlist_value import (
    TracesCreateRequestResourceSpansItemScopeSpansItemSpansItemAttributesItemValueKvlistValue,
)


class TracesCreateRequestResourceSpansItemScopeSpansItemSpansItemAttributesItemValue(
    UniversalBaseModel
):
    string_value: typing_extensions.Annotated[
        typing.Optional[str], FieldMetadata(alias="stringValue")
    ] = None
    int_value: typing_extensions.Annotated[
        typing.Optional[
            TracesCreateRequestResourceSpansItemScopeSpansItemSpansItemAttributesItemValueIntValue
        ],
        FieldMetadata(alias="intValue"),
    ] = None
    double_value: typing_extensions.Annotated[
        typing.Optional[float], FieldMetadata(alias="doubleValue")
    ] = None
    bool_value: typing_extensions.Annotated[
        typing.Optional[bool], FieldMetadata(alias="boolValue")
    ] = None
    array_value: typing_extensions.Annotated[
        typing.Optional[
            TracesCreateRequestResourceSpansItemScopeSpansItemSpansItemAttributesItemValueArrayValue
        ],
        FieldMetadata(alias="arrayValue"),
    ] = None
    kvlist_value: typing_extensions.Annotated[
        typing.Optional[
            TracesCreateRequestResourceSpansItemScopeSpansItemSpansItemAttributesItemValueKvlistValue
        ],
        FieldMetadata(alias="kvlistValue"),
    ] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(
            extra="allow", frozen=True
        )  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



from mirascope import llm


@llm.call("openai/gpt-5-mini")
def recommend_book(genre: str):
    return f"Recommend a {genre} book."


response = recommend_book.stream("fantasy")

try:
    for chunk in response.text_stream():
        print(chunk, end="", flush=True)
except llm.RateLimitError as e:
    print(f"Rate limit exceeded: {e}")
except llm.ConnectionError as e:
    print(f"Connection error: {e}")
except llm.Error as e:
    print(f"Error during streaming: {e}")
